from __future__ import annotations

import re
from typing import Final, Optional, Pattern, Tuple

__all__ = (
    "LAVALINK_BUILD_LINE",
    "LavalinkOldVersion",
    "LavalinkVersion",
)


# present until Lavalink 3.5-rc4
LAVALINK_BUILD_LINE: Final[Pattern] = re.compile(rb"^Build:\s+(?P<build>\d+)$", re.MULTILINE)
# we don't actually care about what the version format before 3.5-rc4 is exactly
# as the comparison is based entirely on the build number
_LAVALINK_VERSION_LINE_PRE35: Final[Pattern] = re.compile(
    rb"^Version:\s+(?P<version>\S+)$", re.MULTILINE | re.VERBOSE
)
# used for LL versions >=3.5-rc4 but below 3.6.
# Since this only applies to historical version, this regex is based only on
# version numbers that actually existed, not ones that technically could.
_LAVALINK_VERSION_LINE_PRE36: Final[Pattern] = re.compile(
    rb"""
    ^
    Version:\s+
    (?P<version>
        (?P<major>3)\.(?P<minor>[0-5])
        # Before LL 3.6, when patch version == 0, it was stripped from the version string
        (?:\.(?P<patch>[1-9]\d*))?
        # Before LL 3.6, the dot in rc.N was optional
        (?:-rc\.?(?P<rc>0|[1-9]\d*))?
        # additional build metadata, can be used by our downstream Lavalink
        # if we need to alter an upstream release
        (?:\+red\.(?P<red>[1-9]\d*))?
    )
    $
    """,
    re.MULTILINE | re.VERBOSE,
)
# used for LL 3.6 and newer
# This regex is limited to the realistic usage in the LL version number,
# not everything that could be a part of it according to the spec.
# We can easily release an update to this regex in the future if it ever becomes necessary.
_LAVALINK_VERSION_LINE: Final[Pattern] = re.compile(
    rb"""
    ^
    Version:\s+
    (?P<version>
        (?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)
        (?:-rc\.(?P<rc>0|[1-9]\d*))?
        # additional build metadata, can be used by our downstream Lavalink
        # if we need to alter an upstream release
        (?:\+red\.(?P<red>[1-9]\d*))?
    )
    $
    """,
    re.MULTILINE | re.VERBOSE,
)


class LavalinkOldVersion:
    def __init__(self, raw_version: str, *, build_number: int) -> None:
        self.raw_version = raw_version
        self.build_number = build_number

    def __str__(self) -> str:
        return f"{self.raw_version}_{self.build_number}"

    @classmethod
    def from_version_output(cls, output: bytes) -> LavalinkOldVersion:
        build_match = LAVALINK_BUILD_LINE.search(output)
        if build_match is None:
            raise ValueError(
                "Could not find 'Build' line in the given `--version` output,"
                " or invalid build number given."
            )
        version_match = _LAVALINK_VERSION_LINE_PRE35.search(output)
        if version_match is None:
            raise ValueError(
                "Could not find 'Version' line in the given `--version` output,"
                " or invalid version number given."
            )
        return cls(
            raw_version=version_match["version"].decode(),
            build_number=int(build_match["build"]),
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LavalinkOldVersion):
            return self.build_number == other.build_number
        if isinstance(other, LavalinkVersion):
            return False
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, LavalinkOldVersion):
            return self.build_number < other.build_number
        if isinstance(other, LavalinkVersion):
            return True
        return NotImplemented

    def __le__(self, other: object) -> bool:
        if isinstance(other, LavalinkOldVersion):
            return self.build_number <= other.build_number
        if isinstance(other, LavalinkVersion):
            return True
        return NotImplemented

    def __gt__(self, other: object) -> bool:
        if isinstance(other, LavalinkOldVersion):
            return self.build_number > other.build_number
        if isinstance(other, LavalinkVersion):
            return False
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        if isinstance(other, LavalinkOldVersion):
            return self.build_number >= other.build_number
        if isinstance(other, LavalinkVersion):
            return False
        return NotImplemented


class LavalinkVersion:
    def __init__(
        self,
        major: int,
        minor: int,
        patch: int = 0,
        *,
        rc: Optional[int] = None,
        red: int = 0,
    ) -> None:
        self.major = major
        self.minor = minor
        self.patch = patch
        self.rc = rc
        self.red = red

    def __str__(self) -> str:
        version = f"{self.major}.{self.minor}.{self.patch}"
        if self.rc is not None:
            version += f"-rc.{self.rc}"
        if self.red:
            version += f"+red.{self.red}"
        return version

    @classmethod
    def from_version_output(cls, output: bytes) -> LavalinkVersion:
        match = _LAVALINK_VERSION_LINE.search(output)
        if match is None:
            # >=3.5-rc4, <3.6
            match = _LAVALINK_VERSION_LINE_PRE36.search(output)
        if match is None:
            raise ValueError(
                "Could not find 'Version' line in the given `--version` output,"
                " or invalid version number given."
            )
        return cls(
            major=int(match["major"]),
            minor=int(match["minor"]),
            patch=int(match["patch"] or 0),
            rc=int(match["rc"]) if match["rc"] is not None else None,
            red=int(match["red"] or 0),
        )

    def _get_comparison_tuple(self) -> Tuple[int, int, int, bool, int, int]:
        return self.major, self.minor, self.patch, self.rc is None, self.rc or 0, self.red

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LavalinkVersion):
            return self._get_comparison_tuple() == other._get_comparison_tuple()
        if isinstance(other, LavalinkOldVersion):
            return False
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, LavalinkVersion):
            return self._get_comparison_tuple() < other._get_comparison_tuple()
        if isinstance(other, LavalinkOldVersion):
            return False
        return NotImplemented

    def __le__(self, other: object) -> bool:
        if isinstance(other, LavalinkVersion):
            return self._get_comparison_tuple() <= other._get_comparison_tuple()
        if isinstance(other, LavalinkOldVersion):
            return False
        return NotImplemented

    def __gt__(self, other: object) -> bool:
        if isinstance(other, LavalinkVersion):
            return self._get_comparison_tuple() > other._get_comparison_tuple()
        if isinstance(other, LavalinkOldVersion):
            return True
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        if isinstance(other, LavalinkVersion):
            return self._get_comparison_tuple() >= other._get_comparison_tuple()
        if isinstance(other, LavalinkOldVersion):
            return True
        return NotImplemented



from redbot.core.commands import BadArgument, Context, Converter
from redbot.core.i18n import Translator

_ = Translator("Cleanup", __file__)

SNOWFLAKE_THRESHOLD = 2**63


class RawMessageIds(Converter):
    async def convert(self, ctx: Context, argument: str) -> int:
        if argument.isnumeric() and len(argument) >= 17 and int(argument) < SNOWFLAKE_THRESHOLD:
            return int(argument)

        raise BadArgument(_("{} doesn't look like a valid message ID.").format(argument))



import sys
import textwrap
from typing import Any, Dict, Optional
from unittest.mock import MagicMock

import pytest

from redbot.core import commands
from redbot.core.dev_commands import DevOutput, SourceCache, cleanup_code


# the examples are based on how the markdown ends up being rendered by Discord
@pytest.mark.parametrize(
    "content,source",
    (
        # no markdown to strip
        (
            "x = 1",
            "x = 1",
        ),
        # inline with single backticks
        (
            "`x = 1`",
            "x = 1",
        ),
        # inline with double backticks
        (
            "``x = 1``",
            "x = 1",
        ),
        # code block within a single line
        (
            "```x = 1```",
            "x = 1",
        ),
        # code block with code in first line and closing backquotes in separate line
        (
            """\
            ```x = 1
            ```""",
            "x = 1",
        ),
        # code block with closing backquotes in same line
        (
            """\
            ```
            x = 1```""",
            "x = 1",
        ),
        # code block with opening and closing backquotes in separate lines
        (
            """\
            ```
            x = 1
            ```""",
            "x = 1",
        ),
        # code block with language specified and closing backquotes in separate line
        (
            """\
            ```py
            x = 1
            ```""",
            "x = 1",
        ),
        (
            """\
            ```python
            x = 1
            ```""",
            "x = 1",
        ),
        # code block with language specified and closing backquotes in same line
        (
            """\
            ```py
            x = 1```""",
            "x = 1",
        ),
        (
            """\
            ```python
            x = 1```""",
            "x = 1",
        ),
        # code block with the only line of code being a potentially valid language name
        # ('pass' is just a combination of letters) and being right after opening backquotes.
        (
            """\
            ```pass
            ```""",
            "pass",
        ),
        # leading newline characters should get stripped, ending backquotes on separate line
        (
            """\
            ```


            x = 1
            ```""",
            "x = 1",
        ),
        (
            """\
            ```python


            x = 1
            ```""",
            "x = 1",
        ),
        # leading newline characters should get stripped, ending backquotes on same line
        (
            """\
            ```


            x = 1```""",
            "x = 1",
        ),
        (
            """\
            ```python


            x = 1```""",
            "x = 1",
        ),
    ),
)
def test_cleanup_code(content: str, source: str) -> None:
    content = textwrap.dedent(content)
    source = textwrap.dedent(source)
    assert cleanup_code(content) == source


def _get_dev_output(
    source: str,
    *,
    source_cache: Optional[SourceCache] = None,
    env: Optional[Dict[str, Any]] = None,
) -> DevOutput:
    if source_cache is None:
        source_cache = SourceCache()
    return DevOutput(
        MagicMock(spec=commands.Context),
        source_cache=source_cache,
        filename=f"<test run - snippet #{source_cache.take_next_index()}>",
        source=source,
        env={"__builtins__": __builtins__, "__name__": "__main__", "_": None, **(env or {})},
    )


async def _run_dev_output(
    monkeypatch: pytest.MonkeyPatch,
    source: str,
    result: str,
    *,
    debug: bool = False,
    eval: bool = False,
    repl: bool = False,
) -> None:
    source = textwrap.dedent(source)
    result = textwrap.dedent(result)
    monkeypatch.setattr("redbot.core.dev_commands.sanitize_output", lambda ctx, s: s)

    if debug:
        output = _get_dev_output(source)
        await output.run_debug()
        assert str(output) == result
        # ensure that our Context mock is never actually used by anything
        assert not output.ctx.mock_calls

    if eval:
        output = _get_dev_output(source.replace("<module>", "func"))
        await output.run_eval()
        assert str(output) == result.replace("<module>", "func")
        # ensure that our Context mock is never actually used by anything
        assert not output.ctx.mock_calls

    if repl:
        output = _get_dev_output(source)
        await output.run_repl()
        assert str(output) == result
        # ensure that our Context mock is never actually used by anything
        assert not output.ctx.mock_calls


EXPRESSION_TESTS = {
    # invalid syntax
    "12x\n": (
        (
            lambda v: v < (3, 10),
            """\
              File "<test run - snippet #0>", line 1
                12x
                  ^
            SyntaxError: invalid syntax
            """,
        ),
        (
            lambda v: v >= (3, 10),
            """\
              File "<test run - snippet #0>", line 1
                12x
                 ^
            SyntaxError: invalid decimal literal
            """,
        ),
    ),
    "foo(x, z for z in range(10), t, w)": (
        (
            lambda v: v < (3, 10),
            """\
              File "<test run - snippet #0>", line 1
                foo(x, z for z in range(10), t, w)
                       ^
            SyntaxError: Generator expression must be parenthesized
            """,
        ),
        (
            lambda v: v >= (3, 10),
            """\
              File "<test run - snippet #0>", line 1
                foo(x, z for z in range(10), t, w)
                       ^^^^^^^^^^^^^^^^^^^^
            SyntaxError: Generator expression must be parenthesized
            """,
        ),
    ),
    # exception raised
    "abs(1 / 0)": (
        (
            lambda v: v < (3, 11),
            """\
            Traceback (most recent call last):
              File "<test run - snippet #0>", line 1, in <module>
                abs(1 / 0)
            ZeroDivisionError: division by zero
            """,
        ),
        (
            lambda v: v >= (3, 11),
            """\
            Traceback (most recent call last):
              File "<test run - snippet #0>", line 1, in <module>
                abs(1 / 0)
                    ~~^~~
            ZeroDivisionError: division by zero
            """,
        ),
    ),
}
STATEMENT_TESTS = {
    # invalid syntax
    """\
    def x():
        12x
    """: (
        (
            lambda v: v < (3, 10),
            """\
              File "<test run - snippet #0>", line 2
                12x
                  ^
            SyntaxError: invalid syntax
            """,
        ),
        (
            lambda v: v >= (3, 10),
            """\
              File "<test run - snippet #0>", line 2
                12x
                 ^
            SyntaxError: invalid decimal literal
            """,
        ),
    ),
    """\
    def x():
        foo(x, z for z in range(10), t, w)
    """: (
        (
            lambda v: v < (3, 10),
            """\
              File "<test run - snippet #0>", line 2
                foo(x, z for z in range(10), t, w)
                       ^
            SyntaxError: Generator expression must be parenthesized
            """,
        ),
        (
            lambda v: v >= (3, 10),
            """\
              File "<test run - snippet #0>", line 2
                foo(x, z for z in range(10), t, w)
                       ^^^^^^^^^^^^^^^^^^^^
            SyntaxError: Generator expression must be parenthesized
            """,
        ),
    ),
    # exception raised
    """\
    print(123)
    try:
        abs(1 / 0)
    except ValueError:
        pass
    """: (
        (
            lambda v: v < (3, 11),
            """\
            123
            Traceback (most recent call last):
              File "<test run - snippet #0>", line 3, in <module>
                abs(1 / 0)
            ZeroDivisionError: division by zero
            """,
        ),
        (
            lambda v: v >= (3, 11),
            """\
            123
            Traceback (most recent call last):
              File "<test run - snippet #0>", line 3, in <module>
                abs(1 / 0)
                    ~~^~~
            ZeroDivisionError: division by zero
            """,
        ),
    ),
    # exception chaining
    """\
    try:
        1 / 0
    except ZeroDivisionError as exc:
        try:
            raise RuntimeError("direct cause") from exc
        except RuntimeError:
            raise ValueError("indirect cause")
    """: (
        (
            lambda v: v < (3, 11),
            """\
            Traceback (most recent call last):
              File "<test run - snippet #0>", line 2, in <module>
                1 / 0
            ZeroDivisionError: division by zero

            The above exception was the direct cause of the following exception:

            Traceback (most recent call last):
              File "<test run - snippet #0>", line 5, in <module>
                raise RuntimeError("direct cause") from exc
            RuntimeError: direct cause

            During handling of the above exception, another exception occurred:

            Traceback (most recent call last):
              File "<test run - snippet #0>", line 7, in <module>
                raise ValueError("indirect cause")
            ValueError: indirect cause
            """,
        ),
        (
            lambda v: v >= (3, 11),
            """\
            Traceback (most recent call last):
              File "<test run - snippet #0>", line 2, in <module>
                1 / 0
                ~~^~~
            ZeroDivisionError: division by zero

            The above exception was the direct cause of the following exception:

            Traceback (most recent call last):
              File "<test run - snippet #0>", line 5, in <module>
                raise RuntimeError("direct cause") from exc
            RuntimeError: direct cause

            During handling of the above exception, another exception occurred:

            Traceback (most recent call last):
              File "<test run - snippet #0>", line 7, in <module>
                raise ValueError("indirect cause")
            ValueError: indirect cause
            """,
        ),
    ),
    # exception groups
    """\
    def f(v):
        try:
            1 / 0
        except ZeroDivisionError:
            try:
                raise ValueError(v)
            except ValueError as e:
                return e
    try:
        raise ExceptionGroup("one", [f(1)])
    except ExceptionGroup as e:
        eg = e
    try:
        raise ExceptionGroup("two", [f(2), eg])
    except ExceptionGroup as e:
        raise RuntimeError("wrapping") from e
    """: (
        (
            lambda v: v >= (3, 11),
            """\
              + Exception Group Traceback (most recent call last):
              |   File "<test run - snippet #0>", line 14, in <module>
              |     raise ExceptionGroup("two", [f(2), eg])
              | ExceptionGroup: two (2 sub-exceptions)
              +-+---------------- 1 ----------------
                | Traceback (most recent call last):
                |   File "<test run - snippet #0>", line 3, in f
                |     1 / 0
                |     ~~^~~
                | ZeroDivisionError: division by zero
                | 
                | During handling of the above exception, another exception occurred:
                | 
                | Traceback (most recent call last):
                |   File "<test run - snippet #0>", line 6, in f
                |     raise ValueError(v)
                | ValueError: 2
                +---------------- 2 ----------------
                | Exception Group Traceback (most recent call last):
                |   File "<test run - snippet #0>", line 10, in <module>
                |     raise ExceptionGroup("one", [f(1)])
                | ExceptionGroup: one (1 sub-exception)
                +-+---------------- 1 ----------------
                  | Traceback (most recent call last):
                  |   File "<test run - snippet #0>", line 3, in f
                  |     1 / 0
                  |     ~~^~~
                  | ZeroDivisionError: division by zero
                  | 
                  | During handling of the above exception, another exception occurred:
                  | 
                  | Traceback (most recent call last):
                  |   File "<test run - snippet #0>", line 6, in f
                  |     raise ValueError(v)
                  | ValueError: 1
                  +------------------------------------

            The above exception was the direct cause of the following exception:

            Traceback (most recent call last):
              File "<test run - snippet #0>", line 16, in <module>
                raise RuntimeError("wrapping") from e
            RuntimeError: wrapping
            """,
        ),
    ),
}


@pytest.mark.parametrize(
    "source,result",
    [
        (source, result)
        for source, results in EXPRESSION_TESTS.items()
        for condition, result in results
        if condition(sys.version_info)
    ],
)
async def test_format_exception_expressions(
    monkeypatch: pytest.MonkeyPatch, source: str, result: str
) -> None:
    await _run_dev_output(monkeypatch, source, result, debug=True, repl=True)


@pytest.mark.parametrize(
    "source,result",
    [
        (source, result)
        for source, results in STATEMENT_TESTS.items()
        for condition, result in results
        if condition(sys.version_info)
    ],
)
async def test_format_exception_statements(
    monkeypatch: pytest.MonkeyPatch, source: str, result: str
) -> None:
    await _run_dev_output(monkeypatch, source, result, eval=True, repl=True)


async def test_successful_run_debug(monkeypatch: pytest.MonkeyPatch) -> None:
    source = "print('hello world'), 123"
    result = "(None, 123)"
    await _run_dev_output(monkeypatch, source, result, debug=True)


async def test_successful_run_eval(monkeypatch: pytest.MonkeyPatch) -> None:
    source = """\
    print("hello world")
    return 123
    """
    result = """\
    hello world
    123"""
    await _run_dev_output(monkeypatch, source, result, eval=True)


async def test_successful_run_repl_eval(monkeypatch: pytest.MonkeyPatch) -> None:
    source = "print('hello world'), 123"
    result = """\
    hello world
    (None, 123)"""
    await _run_dev_output(monkeypatch, source, result, repl=True)


async def test_successful_run_repl_exec(monkeypatch: pytest.MonkeyPatch) -> None:
    source = """\
    print("hello")
    print("world")
    """
    result = """\
    hello
    world
    """
    await _run_dev_output(monkeypatch, source, result, repl=True)


async def test_regression_format_exception_from_previous_snippet(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    snippet_0 = textwrap.dedent(
        """\
    def repro():
        raise Exception("this is an error!")

    return repro
    """
    )
    snippet_1 = "_()"
    result = textwrap.dedent(
        """\
    Traceback (most recent call last):
      File "<test run - snippet #1>", line 1, in func
        _()
      File "<test run - snippet #0>", line 2, in repro
        raise Exception("this is an error!")
    Exception: this is an error!
    """
    )
    monkeypatch.setattr("redbot.core.dev_commands.sanitize_output", lambda ctx, s: s)

    source_cache = SourceCache()
    output = _get_dev_output(snippet_0, source_cache=source_cache)
    await output.run_eval()
    output = _get_dev_output(snippet_1, source_cache=source_cache, env={"_": output.result})
    await output.run_eval()
    assert str(output) == result
    # ensure that our Context mock is never actually used by anything
    assert not output.ctx.mock_calls



"""Errors module for the app_commands package."""
from discord import app_commands


class UserFeedbackCheckFailure(app_commands.CheckFailure):
    """A version of CheckFailure responding with a custom error message."""

    def __init__(self, message=None, *args):
        self.message = message
        super().__init__(message, *args)



import datetime
import json
from collections import namedtuple
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, MutableMapping, Optional, Union

import discord
import lavalink
from red_commons.logging import getLogger

from redbot.core.bot import Red
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import humanize_list

from ..errors import InvalidPlaylistScope, MissingAuthor, MissingGuild
from ..utils import PlaylistScope

log = getLogger("red.cogs.Audio.api.utils")
_ = Translator("Audio", Path(__file__))


@dataclass
class YouTubeCacheFetchResult:
    query: Optional[str]
    last_updated: int

    def __post_init__(self):
        if isinstance(self.last_updated, int):
            self.updated_on: datetime.datetime = datetime.datetime.fromtimestamp(self.last_updated)


@dataclass
class SpotifyCacheFetchResult:
    query: Optional[str]
    last_updated: int

    def __post_init__(self):
        if isinstance(self.last_updated, int):
            self.updated_on: datetime.datetime = datetime.datetime.fromtimestamp(self.last_updated)


@dataclass
class LavalinkCacheFetchResult:
    query: Optional[MutableMapping]
    last_updated: int

    def __post_init__(self):
        if isinstance(self.last_updated, int):
            self.updated_on: datetime.datetime = datetime.datetime.fromtimestamp(self.last_updated)

        if isinstance(self.query, str):
            self.query = json.loads(self.query)


@dataclass
class LavalinkCacheFetchForGlobalResult:
    query: str
    data: MutableMapping

    def __post_init__(self):
        if isinstance(self.data, str):
            self.data_string = str(self.data)
            self.data = json.loads(self.data)


@dataclass
class PlaylistFetchResult:
    playlist_id: int
    playlist_name: str
    scope_id: int
    author_id: int
    playlist_url: Optional[str] = None
    tracks: List[MutableMapping] = field(default_factory=lambda: [])

    def __post_init__(self):
        if isinstance(self.tracks, str):
            self.tracks = json.loads(self.tracks)


@dataclass
class QueueFetchResult:
    guild_id: int
    room_id: int
    track: dict = field(default_factory=lambda: {})
    track_object: lavalink.Track = None

    def __post_init__(self):
        if isinstance(self.track, str):
            self.track = json.loads(self.track)
        if self.track:
            self.track_object = lavalink.Track(self.track)


def standardize_scope(scope: str) -> str:
    """Convert any of the used scopes into one we are expecting."""
    scope = scope.upper()
    valid_scopes = ["GLOBAL", "GUILD", "AUTHOR", "USER", "SERVER", "MEMBER", "BOT"]

    if scope in PlaylistScope.list():
        return scope
    elif scope not in valid_scopes:
        raise InvalidPlaylistScope(
            f'"{scope}" is not a valid playlist scope.'
            f" Scope needs to be one of the following: {humanize_list(valid_scopes)}"
        )

    if scope in ["GLOBAL", "BOT"]:
        scope = PlaylistScope.GLOBAL.value
    elif scope in ["GUILD", "SERVER"]:
        scope = PlaylistScope.GUILD.value
    elif scope in ["USER", "MEMBER", "AUTHOR"]:
        scope = PlaylistScope.USER.value

    return scope


def prepare_config_scope(
    bot: Red,
    scope,
    author: Union[discord.abc.User, int] = None,
    guild: Union[discord.Guild, int] = None,
):
    """Return the scope used by Playlists."""
    scope = standardize_scope(scope)
    if scope == PlaylistScope.GLOBAL.value:
        config_scope = [PlaylistScope.GLOBAL.value, bot.user.id]
    elif scope == PlaylistScope.USER.value:
        if author is None:
            raise MissingAuthor("Invalid author for user scope.")
        config_scope = [PlaylistScope.USER.value, int(getattr(author, "id", author))]
    else:
        if guild is None:
            raise MissingGuild("Invalid guild for guild scope.")
        config_scope = [PlaylistScope.GUILD.value, int(getattr(guild, "id", guild))]
    return config_scope


def prepare_config_scope_for_migration23(  # TODO: remove me in a future version ?
    scope, author: Union[discord.abc.User, int] = None, guild: discord.Guild = None
):
    """Return the scope used by Playlists."""
    scope = standardize_scope(scope)

    if scope == PlaylistScope.GLOBAL.value:
        config_scope = [PlaylistScope.GLOBAL.value]
    elif scope == PlaylistScope.USER.value:
        if author is None:
            raise MissingAuthor("Invalid author for user scope.")
        config_scope = [PlaylistScope.USER.value, str(getattr(author, "id", author))]
    else:
        if guild is None:
            raise MissingGuild("Invalid guild for guild scope.")
        config_scope = [PlaylistScope.GUILD.value, str(getattr(guild, "id", guild))]
    return config_scope


FakePlaylist = namedtuple("Playlist", "author scope")



from __future__ import annotations

from typing import Any, Dict, Final

from . import version_pins

__all__ = (
    "DEFAULT_LAVALINK_YAML",
    "get_default_server_config",
    "generate_server_config",
    "change_dict_naming_convention",
)

YT_PLUGIN_REPOSITORY: Final[str] = "https://maven.lavalink.dev/releases"

DEFAULT_LAVALINK_YAML = {
    # The nesting structure of this dict is very important, it's a 1:1 mirror of application.yaml in JSON
    "yaml__server__address": "localhost",
    "yaml__server__port": 2333,
    "yaml__lavalink__server__password": "youshallnotpass",
    "yaml__lavalink__server__sources__http": True,
    "yaml__lavalink__server__sources__bandcamp": True,
    "yaml__lavalink__server__sources__local": True,
    "yaml__lavalink__server__sources__soundcloud": True,
    "yaml__lavalink__server__sources__youtube": True,
    "yaml__lavalink__server__sources__twitch": True,
    "yaml__lavalink__server__sources__vimeo": True,
    "yaml__lavalink__server__bufferDurationMs": 400,
    "yaml__lavalink__server__frameBufferDurationMs": 1000,
    # 100 pages - 100 entries per page = 10,000 tracks which is the Audio Limit for a single playlist.
    "yaml__lavalink__server__youtubePlaylistLoadLimit": 100,
    "yaml__lavalink__server__playerUpdateInterval": 1,
    "yaml__lavalink__server__youtubeSearchEnabled": True,
    "yaml__lavalink__server__soundcloudSearchEnabled": True,
    "yaml__lavalink__server__gc_warnings": True,
    "yaml__metrics__prometheus__enabled": False,
    "yaml__metrics__prometheus__endpoint": "/metrics",
    "yaml__sentry__dsn": "",
    "yaml__sentry__environment": "",
    "yaml__logging__file__path": "./logs/",
    "yaml__logging__level__root": "INFO",
    "yaml__logging__level__lavalink": "INFO",
    "yaml__logging__logback__rollingpolicy__max_history": 15,
    "yaml__logging__logback__rollingpolicy__max_size": "10MB",
    # plugin configuration - note that the plugin may be disabled by the manager
    "yaml__plugins__youtube__enabled": True,
    "yaml__plugins__youtube__allowSearch": True,
    "yaml__plugins__youtube__allowDirectVideoIds": True,
    "yaml__plugins__youtube__allowDirectPlaylistIds": True,
    "yaml__plugins__youtube__clients": [
        "MUSIC",
        "WEB",
        "WEBEMBEDDED",
        "MWEB",
        "TVHTML5EMBEDDED",
        "TV",
        "IOS",
    ],
    "yaml__plugins__youtube__MUSIC__playback": False,
    "yaml__plugins__youtube__MUSIC__playlistLoading": False,
    "yaml__plugins__youtube__MUSIC__searching": True,
    "yaml__plugins__youtube__MUSIC__videoLoading": False,
    "yaml__plugins__youtube__MWEB__playback": True,
    "yaml__plugins__youtube__MWEB__playlistLoading": True,
    "yaml__plugins__youtube__MWEB__searching": False,
    "yaml__plugins__youtube__MWEB__videoLoading": False,
    "yaml__plugins__youtube__TV__playback": True,
    "yaml__plugins__youtube__TV__playlistLoading": False,
    "yaml__plugins__youtube__TV__searching": False,
    "yaml__plugins__youtube__TV__videoLoading": False,
    "yaml__plugins__youtube__TVHTML5EMBEDDED__playback": True,
    "yaml__plugins__youtube__TVHTML5EMBEDDED__playlistLoading": False,
    "yaml__plugins__youtube__TVHTML5EMBEDDED__searching": False,
    "yaml__plugins__youtube__TVHTML5EMBEDDED__videoLoading": False,
    "yaml__plugins__youtube__WEB__playback": True,
    "yaml__plugins__youtube__WEB__playlistLoading": True,
    "yaml__plugins__youtube__WEB__searching": False,
    "yaml__plugins__youtube__WEB__videoLoading": False,
    "yaml__plugins__youtube__WEBEMBEDDED__playback": True,
    "yaml__plugins__youtube__WEBEMBEDDED__playlistLoading": False,
    "yaml__plugins__youtube__WEBEMBEDDED__searching": False,
    "yaml__plugins__youtube__WEBEMBEDDED__videoLoading": False,
}


def _unflatten_config_defaults(config_defaults: Dict[str, Any]) -> Dict[str, Any]:
    ret: Dict[str, Any] = {}

    # based on Config._get_defaults_dict()
    for flat_key, value in config_defaults.items():
        keys = flat_key.split("__")
        partial = ret
        for idx, key in enumerate(keys, start=1):
            if idx == len(keys):
                partial[key] = value
            else:
                partial = partial.setdefault(key, {})

    return ret


def get_default_server_config() -> Dict[str, Any]:
    return generate_server_config(_unflatten_config_defaults(DEFAULT_LAVALINK_YAML)["yaml"])


def generate_server_config(config_data: Dict[str, Any]) -> Dict[str, Any]:
    data = change_dict_naming_convention(config_data)
    ll_config = data["lavalink"]
    sources = ll_config["server"]["sources"]
    plugins = ll_config.setdefault("plugins", [])

    enable_yt_plugin = sources["youtube"]
    if enable_yt_plugin:
        sources["youtube"] = False
        yt_plugin = {
            "dependency": f"dev.lavalink.youtube:youtube-plugin:{version_pins.YT_PLUGIN_VERSION}",
            "repository": YT_PLUGIN_REPOSITORY,
        }
        plugins.append(yt_plugin)

    return data


# This assumes all keys with `_` should be converted from `part1_part2` to `part1-part2`
def _convert_function(key: str) -> str:
    return key.replace("_", "-")


def change_dict_naming_convention(data: Any) -> Any:
    ret: Any = data
    if isinstance(data, dict):
        ret = {}
        for key, value in data.items():
            ret[_convert_function(key)] = change_dict_naming_convention(value)
    elif isinstance(data, list):
        ret = []
        for value in data:
            ret.append(change_dict_naming_convention(value))
    return ret



import asyncio
import contextlib
import datetime
from pathlib import Path
from typing import Dict

import discord
import lavalink
from discord.backoff import ExponentialBackoff
from red_commons.logging import getLogger

from redbot.core.i18n import Translator, set_contextual_locales_from_guild
from ...errors import DatabaseError, TrackEnqueueError
from ..abc import MixinMeta
from ..cog_utils import CompositeMetaClass

log = getLogger("red.cogs.Audio.cog.Events.lavalink")
ws_audio_log = getLogger("red.Audio.WS.Audio")

_ = Translator("Audio", Path(__file__))


class LavalinkEvents(MixinMeta, metaclass=CompositeMetaClass):
    async def lavalink_update_handler(
        self, player: lavalink.Player, event_type: lavalink.enums.PlayerState, extra
    ):
        self._last_ll_update = datetime.datetime.now(datetime.timezone.utc)
        self._ll_guild_updates.add(int(extra.get("guildId", 0)))

    async def lavalink_event_handler(
        self, player: lavalink.Player, event_type: lavalink.LavalinkEvents, extra
    ) -> None:
        current_track = player.current
        current_channel = player.channel
        guild = self.rgetattr(current_channel, "guild", None)
        if not (current_channel and guild):
            player.store("autoplay_notified", False)
            await player.stop()
            await player.disconnect()
            return
        if await self.bot.cog_disabled_in_guild(self, guild):
            await player.stop()
            await player.disconnect()
            if guild:
                await self.config.guild_from_id(guild_id=guild.id).currently_auto_playing_in.set(
                    []
                )
            return
        guild_id = self.rgetattr(guild, "id", None)
        if not guild:
            return
        # This event is rather spammy during playback - specially if there's multiple player
        #  Lets move it to Verbose that way it still there if needed alongside the other more verbose content.
        guild_data = await self.config.guild(guild).all()
        disconnect = guild_data["disconnect"]
        if event_type == lavalink.LavalinkEvents.FORCED_DISCONNECT:
            self.bot.dispatch("red_audio_audio_disconnect", guild)
            self._ll_guild_updates.discard(guild.id)
            return
        if event_type == lavalink.LavalinkEvents.WEBSOCKET_CLOSED:
            deafen = guild_data["auto_deafen"]
            event_channel_id = extra.get("channelID")
            _error_code = extra.get("code")
            if _error_code in [1000] or not guild:
                if _error_code == 1000 and player.current is not None and player.is_playing:
                    await player.resume(player.current, start=player.position, replace=True)
                    by_remote = extra.get("byRemote", "")
                    reason = extra.get("reason", "No Specified Reason").strip()
                    ws_audio_log.info(
                        "WS EVENT - SIMPLE RESUME (Healthy Socket) | "
                        "Voice websocket closed event "
                        "Code: %s -- Remote: %s -- %s",
                        extra.get("code"),
                        by_remote,
                        reason,
                    )
                    ws_audio_log.debug(
                        "WS EVENT - SIMPLE RESUME (Healthy Socket) | "
                        "Voice websocket closed event "
                        "Code: %s -- Remote: %s -- %s, %r",
                        extra.get("code"),
                        by_remote,
                        reason,
                        player,
                    )
                return
            await self._ws_op_codes[guild_id].put((event_channel_id, _error_code))
            try:
                if guild_id not in self._ws_resume:
                    self._ws_resume[guild_id].set()

                await self._websocket_closed_handler(
                    guild=guild,
                    player=player,
                    extra=extra,
                    self_deaf=deafen,
                    disconnect=disconnect,
                )
            except Exception as exc:
                log.debug(
                    "Error in WEBSOCKET_CLOSED handling for guild: %s",
                    player.guild.id,
                    exc_info=exc,
                )
            return
        if not player.node.ready:
            log.debug("Player node is not ready discarding event")
            log.verbose(
                "Received a new discard lavalink event for %s: %s: %r", guild_id, event_type, extra
            )
            return
        log.verbose("Received a new lavalink event for %s: %s: %r", guild_id, event_type, extra)
        await set_contextual_locales_from_guild(self.bot, guild)
        current_requester = self.rgetattr(current_track, "requester", None)
        current_stream = self.rgetattr(current_track, "is_stream", None)
        current_length = self.rgetattr(current_track, "length", None)
        current_thumbnail = self.rgetattr(current_track, "thumbnail", None)
        current_id = self.rgetattr(current_track, "_info", {}).get("identifier")

        repeat = guild_data["repeat"]
        notify = guild_data["notify"]
        autoplay = guild_data["auto_play"]
        description = await self.get_track_description(
            current_track, self.local_folder_current_path
        )
        status = await self.config.status()
        prev_song: lavalink.Track = player.fetch("prev_song")
        await self.maybe_reset_error_counter(player)

        if event_type == lavalink.LavalinkEvents.TRACK_START:
            self.skip_votes[guild_id] = set()
            playing_song = player.fetch("playing_song")
            requester = player.fetch("requester")
            player.store("prev_song", playing_song)
            player.store("prev_requester", requester)
            player.store("playing_song", current_track)
            player.store("requester", current_requester)
            self.bot.dispatch("red_audio_track_start", guild, current_track, current_requester)
            if guild_id and current_track:
                await self.api_interface.persistent_queue_api.played(
                    guild_id=guild_id, track_id=current_track.track_identifier
                )
            notify_channel = player.fetch("notify_channel")
            if notify_channel and autoplay:
                await self.config.guild_from_id(guild_id=guild_id).currently_auto_playing_in.set(
                    [notify_channel, player.channel.id]
                )
            else:
                await self.config.guild_from_id(guild_id=guild_id).currently_auto_playing_in.set(
                    []
                )
        if event_type == lavalink.LavalinkEvents.TRACK_END:
            prev_requester = player.fetch("prev_requester")
            self.bot.dispatch("red_audio_track_end", guild, prev_song, prev_requester)
            player.store("resume_attempts", 0)
        if event_type == lavalink.LavalinkEvents.QUEUE_END:
            prev_requester = player.fetch("prev_requester")
            self.bot.dispatch("red_audio_queue_end", guild, prev_song, prev_requester)
            if guild_id:
                await self.api_interface.persistent_queue_api.drop(guild_id)
            if player.is_auto_playing or (
                autoplay
                and not player.queue
                and player.fetch("playing_song") is not None
                and self.playlist_api is not None
                and self.api_interface is not None
            ):
                notify_channel_id = player.fetch("notify_channel")
                try:
                    await self.api_interface.autoplay(player, self.playlist_api)
                except DatabaseError:
                    notify_channel = guild.get_channel_or_thread(notify_channel_id)
                    if notify_channel and self._has_notify_perms(notify_channel):
                        await self.send_embed_msg(
                            notify_channel, title=_("Couldn't get a valid track.")
                        )
                    return
                except TrackEnqueueError:
                    notify_channel = guild.get_channel_or_thread(notify_channel_id)
                    if notify_channel and self._has_notify_perms(notify_channel):
                        await self.send_embed_msg(
                            notify_channel,
                            title=_("Unable to Get Track"),
                            description=_(
                                "I'm unable to get a track from the Lavalink node at the moment, try again in a few "
                                "minutes."
                            ),
                        )
                    return
        if event_type == lavalink.LavalinkEvents.TRACK_START and notify:
            notify_channel_id = player.fetch("notify_channel")
            notify_channel = guild.get_channel_or_thread(notify_channel_id)
            if notify_channel and self._has_notify_perms(notify_channel):
                if player.fetch("notify_message") is not None:
                    with contextlib.suppress(discord.HTTPException):
                        await player.fetch("notify_message").delete()
                if not (description and notify_channel):
                    return
                if current_stream:
                    dur = "LIVE"
                else:
                    dur = self.format_time(current_length)

                thumb = None
                if await self.config.guild(guild).thumbnail() and current_thumbnail:
                    thumb = current_thumbnail

                notify_message = await self.send_embed_msg(
                    notify_channel,
                    title=_("Now Playing"),
                    description=description,
                    footer=_("Track length: {length} | Requested by: {user}").format(
                        length=dur, user=current_requester
                    ),
                    thumbnail=thumb,
                )
                player.store("notify_message", notify_message)
        if event_type == lavalink.LavalinkEvents.TRACK_START and status:
            log.debug("Track started for %s, updating bot status", guild_id)
            player_check = await self.get_active_player_count()
            await self.update_bot_presence(*player_check)

        if event_type == lavalink.LavalinkEvents.TRACK_END and status:
            await asyncio.sleep(1)
            if not player.is_playing:
                log.debug("Track ended for %s, updating bot status", guild_id)
                player_check = await self.get_active_player_count()
                await self.update_bot_presence(*player_check)

        if event_type == lavalink.LavalinkEvents.QUEUE_END:
            if not autoplay:
                notify_channel_id = player.fetch("notify_channel")
                notify_channel = guild.get_channel_or_thread(notify_channel_id)
                if notify_channel and notify and self._has_notify_perms(notify_channel):
                    await self.send_embed_msg(notify_channel, title=_("Queue ended."))
                if disconnect:
                    log.debug(
                        "Queue ended for %s, Disconnecting bot due to configuration", guild_id
                    )
                    self.bot.dispatch("red_audio_audio_disconnect", guild)
                    await self.config.guild_from_id(
                        guild_id=guild_id
                    ).currently_auto_playing_in.set([])
                    # let audio buffer run out on slower machines (GH-5158)
                    await asyncio.sleep(2)
                    await player.disconnect()
                    self._ll_guild_updates.discard(guild.id)
            if status:
                log.debug("Queue ended for %s, updating bot status", guild_id)
                player_check = await self.get_active_player_count()
                await self.update_bot_presence(*player_check)

        if event_type in [
            lavalink.LavalinkEvents.TRACK_EXCEPTION,
            lavalink.LavalinkEvents.TRACK_STUCK,
        ]:
            message_channel = player.fetch("notify_channel")
            while True:
                if current_track in player.queue:
                    player.queue.remove(current_track)
                else:
                    break
            if repeat:
                player.current = None
            if not guild_id:
                return
            guild_id = int(guild_id)
            self._error_counter.setdefault(guild_id, 0)
            if guild_id not in self._error_counter:
                self._error_counter[guild_id] = 0
            early_exit = await self.increase_error_counter(player)
            if early_exit:
                self._disconnected_players[guild_id] = True
                self.play_lock[guild_id] = False
                eq = player.fetch("eq")
                player.queue = []
                player.store("playing_song", None)
                player.store("autoplay_notified", False)
                if eq:
                    await self.config.custom("EQUALIZER", guild_id).eq_bands.set(eq.bands)
                await player.stop()
                await player.disconnect()
                await self.config.guild_from_id(guild_id=guild_id).currently_auto_playing_in.set(
                    []
                )
                self._ll_guild_updates.discard(guild_id)
                self.bot.dispatch("red_audio_audio_disconnect", guild)
            if message_channel:
                message_channel = guild.get_channel_or_thread(message_channel)
                if early_exit:
                    log.warning(
                        "Audio detected multiple continuous errors during playback "
                        "- terminating the player for guild: %s.",
                        guild_id,
                    )
                    log.verbose(
                        "Player has been terminated due to multiple playback failures: %r", player
                    )
                    embed = discord.Embed(
                        colour=await self.bot.get_embed_color(message_channel),
                        title=_("Multiple Errors Detected"),
                        description=_(
                            "Closing the audio player "
                            "due to multiple errors being detected. "
                            "If this persists, please inform the bot owner "
                            "as the Audio cog may be temporally unavailable."
                        ),
                    )
                    await message_channel.send(embed=embed)
                    return
                else:
                    description = description or ""
                    if event_type == lavalink.LavalinkEvents.TRACK_STUCK:
                        embed = discord.Embed(
                            colour=await self.bot.get_embed_color(message_channel),
                            title=_("Track Stuck"),
                            description=_(
                                "Playback of the song has stopped due to an unexpected error.\n{error}"
                            ).format(error=description),
                        )
                    else:
                        embed = discord.Embed(
                            title=_("Track Error"),
                            colour=await self.bot.get_embed_color(message_channel),
                            description="{}\n{}".format(
                                extra["message"].replace("\n", ""), description
                            ),
                        )
                        if current_id:
                            asyncio.create_task(
                                self.api_interface.global_cache_api.report_invalid(current_id)
                            )
                    await message_channel.send(embed=embed)
            if player.node.ready:
                await player.skip()

    async def _websocket_closed_handler(
        self,
        guild: discord.Guild,
        player: lavalink.Player,
        extra: Dict,
        self_deaf: bool,
        disconnect: bool,
    ) -> None:
        guild_id = guild.id
        shard = self.bot.shards[guild.shard_id]
        event_channel_id = extra.get("channelID")
        try:
            if not self._ws_resume[guild_id].is_set():
                await self._ws_resume[guild_id].wait()
            else:
                self._ws_resume[guild_id].clear()
            code = extra.get("code")
            by_remote = extra.get("byRemote", "")
            reason = extra.get("reason", "No Specified Reason").strip()
            channel_id = player.channel.id
            try:
                event_channel_id, to_handle_code = await self._ws_op_codes[guild_id].get()
            except asyncio.QueueEmpty:
                log.debug("Empty queue - Resuming Processor - Early exit")
                return

            if code != to_handle_code:
                code = to_handle_code
                if player.channel.id != event_channel_id:
                    code = 4014
            if event_channel_id != channel_id:
                ws_audio_log.debug(
                    "Received an op code for a channel that is no longer valid; %s "
                    "Reason: Error code %s & %s, %r",
                    event_channel_id,
                    code,
                    reason,
                    player,
                )
                self._ws_op_codes[guild_id]._init(self._ws_op_codes[guild_id]._maxsize)
                return
            if player.channel:
                has_perm = self.can_join_and_speak(player.channel)
            else:
                has_perm = False
            if code in (1000,) and has_perm and player.current and player.is_playing:
                player.store("resumes", player.fetch("resumes", 0) + 1)
                await player.resume(player.current, start=player.position, replace=True)
                ws_audio_log.info("Player resumed | Reason: Error code %s & %s", code, reason)
                ws_audio_log.debug(
                    "Player resumed | Reason: Error code %s & %s, %r", code, reason, player
                )
                self._ws_op_codes[guild_id]._init(self._ws_op_codes[guild_id]._maxsize)
                return

            if shard.is_closed():
                if player._con_delay:
                    delay = player._con_delay.delay()
                else:
                    player._con_delay = ExponentialBackoff(base=1)
                    delay = player._con_delay.delay()
                ws_audio_log.debug(
                    "YOU CAN IGNORE THIS UNLESS IT'S CONSISTENTLY REPEATING FOR THE SAME GUILD - "
                    "Voice websocket closed for guild %s -> "
                    "Socket Closed %s.  "
                    "Code: %s -- Remote: %s -- %s, %r",
                    guild_id,
                    shard.is_closed(),
                    code,
                    by_remote,
                    reason,
                    player,
                )
                ws_audio_log.info(
                    "Reconnecting to channel %s in guild: %s | %.2fs",
                    channel_id,
                    guild_id,
                    delay,
                )
                await asyncio.sleep(delay)
                while shard.is_closed():
                    await asyncio.sleep(0.1)

                if has_perm and player.current and player.is_playing:
                    player.store("resumes", player.fetch("resumes", 0) + 1)
                    await player.connect(self_deaf=self_deaf)
                    await player.resume(player.current, start=player.position, replace=True)
                    ws_audio_log.info(
                        "Voice websocket reconnected Reason: Error code %s & Currently playing",
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket reconnected "
                        "Reason: Error code %s & Currently playing, %r",
                        code,
                        player,
                    )
                elif has_perm and player.paused and player.current:
                    player.store("resumes", player.fetch("resumes", 0) + 1)
                    await player.connect(self_deaf=self_deaf)
                    await player.resume(
                        player.current, start=player.position, replace=True, pause=True
                    )
                    ws_audio_log.info(
                        "Voice websocket reconnected Reason: Error code %s & Currently Paused",
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket reconnected "
                        "Reason: Error code %s & Currently Paused, %r",
                        code,
                        player,
                    )
                elif has_perm and (not disconnect) and (not player.is_playing):
                    player.store("resumes", player.fetch("resumes", 0) + 1)
                    await player.connect(self_deaf=self_deaf)
                    ws_audio_log.info(
                        "Voice websocket reconnected "
                        "Reason: Error code %s & Not playing, but auto disconnect disabled",
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket reconnected "
                        "Reason: Error code %s & Not playing, but auto disconnect disabled, %r",
                        code,
                        player,
                    )
                    self._ll_guild_updates.discard(guild_id)
                elif not has_perm:
                    self.bot.dispatch("red_audio_audio_disconnect", guild)
                    ws_audio_log.info(
                        "Voice websocket disconnected "
                        "Reason: Error code %s & Missing permissions",
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket disconnected "
                        "Reason: Error code %s & Missing permissions, %r",
                        code,
                        player,
                    )
                    self._ll_guild_updates.discard(guild_id)
                    player.store("autoplay_notified", False)
                    await player.stop()
                    await player.disconnect()
                    await self.config.guild_from_id(
                        guild_id=guild_id
                    ).currently_auto_playing_in.set([])
                else:
                    self.bot.dispatch("red_audio_audio_disconnect", guild)
                    ws_audio_log.info(
                        "Voice websocket disconnected Reason: Error code %s & Unknown", code
                    )
                    ws_audio_log.debug(
                        "Voice websocket disconnected Reason: Error code %s & Unknown, %r",
                        code,
                        player,
                    )
                    self._ll_guild_updates.discard(guild_id)
                    player.store("autoplay_notified", False)
                    await player.stop()
                    await player.disconnect()
                    await self.config.guild_from_id(
                        guild_id=guild_id
                    ).currently_auto_playing_in.set([])
            elif code in (42069,) and has_perm and player.current and player.is_playing:
                player.store("resumes", player.fetch("resumes", 0) + 1)
                await player.connect(self_deaf=self_deaf)
                await player.resume(player.current, start=player.position, replace=True)
                ws_audio_log.info("Player resumed - Reason: Error code %s & %s", code, reason)
                ws_audio_log.debug(
                    "Player resumed - Reason: Error code %s & %s, %r", code, reason, player
                )
            elif code in (4015, 4009, 4006, 4000, 1006):
                if player._con_delay:
                    delay = player._con_delay.delay()
                else:
                    player._con_delay = ExponentialBackoff(base=1)
                    delay = player._con_delay.delay()
                ws_audio_log.debug(
                    "Reconnecting to channel %s in guild: %s | %.2fs", channel_id, guild_id, delay
                )
                await asyncio.sleep(delay)
                if has_perm and player.current and player.is_playing:
                    await player.connect(self_deaf=self_deaf)
                    await player.resume(player.current, start=player.position, replace=True)
                    ws_audio_log.info(
                        "Voice websocket reconnected Reason: Error code %s & Player is active",
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket reconnected "
                        "Reason: Error code %s & Player is active, %r",
                        code,
                        player,
                    )
                elif has_perm and player.paused and player.current:
                    player.store("resumes", player.fetch("resumes", 0) + 1)
                    await player.connect(self_deaf=self_deaf)
                    await player.resume(
                        player.current, start=player.position, replace=True, pause=True
                    )
                    ws_audio_log.info(
                        "Voice websocket reconnected Reason: Error code %s & Player is paused",
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket reconnected "
                        "Reason: Error code %s & Player is paused, %r",
                        code,
                        player,
                    )
                elif has_perm and (not disconnect) and (not player.is_playing):
                    player.store("resumes", player.fetch("resumes", 0) + 1)
                    await player.connect(self_deaf=self_deaf)
                    ws_audio_log.info(
                        "Voice websocket reconnected "
                        "to channel %s in guild: %s | "
                        "Reason: Error code %s & Not playing",
                        channel_id,
                        guild_id,
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket reconnected "
                        "to channel %s in guild: %s | "
                        "Reason: Error code %s & Not playing, %r",
                        channel_id,
                        guild_id,
                        code,
                        player,
                    )
                    self._ll_guild_updates.discard(guild_id)
                elif not has_perm:
                    self.bot.dispatch("red_audio_audio_disconnect", guild)
                    ws_audio_log.info(
                        "Voice websocket disconnected "
                        "Reason: Error code %s & Missing permissions",
                        code,
                    )
                    ws_audio_log.debug(
                        "Voice websocket disconnected "
                        "Reason: Error code %s & Missing permissions, %r",
                        code,
                        player,
                    )
                    self._ll_guild_updates.discard(guild_id)
                    player.store("autoplay_notified", False)
                    await player.stop()
                    await player.disconnect()
                    await self.config.guild_from_id(
                        guild_id=guild_id
                    ).currently_auto_playing_in.set([])
            else:
                if not player.paused and player.current:
                    player.store("resumes", player.fetch("resumes", 0) + 1)
                    await player.resume(player.current, start=player.position, replace=True)
                    ws_audio_log.info(
                        "WS EVENT - SIMPLE RESUME (Healthy Socket) | "
                        "Voice websocket closed event "
                        "Code: %s -- Remote: %s -- %s",
                        code,
                        by_remote,
                        reason,
                    )
                    ws_audio_log.debug(
                        "WS EVENT - SIMPLE RESUME (Healthy Socket) | "
                        "Voice websocket closed event "
                        "Code: %s -- Remote: %s -- %s, %r",
                        code,
                        by_remote,
                        reason,
                        player,
                    )
                else:
                    ws_audio_log.info(
                        "WS EVENT - IGNORED (Healthy Socket) | "
                        "Voice websocket closed event "
                        "Code: %s -- Remote: %s -- %s",
                        code,
                        by_remote,
                        reason,
                    )
                    ws_audio_log.debug(
                        "WS EVENT - IGNORED (Healthy Socket) | "
                        "Voice websocket closed event "
                        "Code: %s -- Remote: %s -- %s, %r",
                        code,
                        by_remote,
                        reason,
                        player,
                    )
        except Exception as exc:
            log.exception("Error in task", exc_info=exc)
        finally:
            self._ws_op_codes[guild_id]._init(self._ws_op_codes[guild_id]._maxsize)
            self._ws_resume[guild_id].set()



import asyncio
import pytest
import operator
import random
from redbot.core.utils import (
    bounded_gather,
    bounded_gather_iter,
    deduplicate_iterables,
    common_filters,
)
from redbot.core.utils.chat_formatting import pagify
from typing import List


def test_deduplicate_iterables():
    expected = [1, 2, 3, 4, 5]
    inputs = [[1, 2, 1], [3, 1, 2, 4], [5, 1, 2]]
    assert deduplicate_iterables(*inputs) == expected


async def test_bounded_gather():
    status = [0, 0]  # num_running, max_running

    async def wait_task(i, delay, status, fail=False):
        status[0] += 1
        await asyncio.sleep(delay)
        status[1] = max(status)
        status[0] -= 1

        if fail:
            raise RuntimeError

        return i

    num_concurrent = random.randint(2, 8)
    num_tasks = random.randint(4 * num_concurrent, 5 * num_concurrent)
    num_fail = random.randint(num_concurrent, num_tasks)

    tasks = [wait_task(i, random.random() / 1000, status) for i in range(num_tasks)]
    tasks += [wait_task(i, random.random() / 1000, status, fail=True) for i in range(num_fail)]

    num_failed = 0

    results = await bounded_gather(*tasks, limit=num_concurrent, return_exceptions=True)

    for i, result in enumerate(results):
        if isinstance(result, RuntimeError):
            num_failed += 1
        else:
            assert result == i  # verify_permissions original order
            assert 0 <= result < num_tasks

    assert 0 < status[1] <= num_concurrent
    assert num_fail == num_failed


async def test_bounded_gather_iter():
    status = [0, 0]  # num_running, max_running

    async def wait_task(i, delay, status, fail=False):
        status[0] += 1
        await asyncio.sleep(delay)
        status[1] = max(status)
        status[0] -= 1

        if fail:
            raise RuntimeError

        return i

    num_concurrent = random.randint(2, 8)
    num_tasks = random.randint(4 * num_concurrent, 16 * num_concurrent)
    num_fail = random.randint(num_concurrent, num_tasks)

    tasks = [wait_task(i, random.random() / 1000, status) for i in range(num_tasks)]
    tasks += [wait_task(i, random.random() / 1000, status, fail=True) for i in range(num_fail)]
    random.shuffle(tasks)

    num_failed = 0

    for result in bounded_gather_iter(*tasks, limit=num_concurrent):
        try:
            result = await result
        except RuntimeError:
            num_failed += 1
            continue

        assert 0 <= result < num_tasks

    assert 0 < status[1] <= num_concurrent
    assert num_fail == num_failed


@pytest.mark.skip(reason="spams logs with pending task warnings")
async def test_bounded_gather_iter_cancel():
    status = [0, 0, 0]  # num_running, max_running, num_ran

    async def wait_task(i, delay, status, fail=False):
        status[0] += 1
        await asyncio.sleep(delay)
        status[1] = max(status[:2])
        status[0] -= 1

        if fail:
            raise RuntimeError

        status[2] += 1
        return i

    num_concurrent = random.randint(2, 8)
    num_tasks = random.randint(4 * num_concurrent, 16 * num_concurrent)
    quit_on = random.randint(0, num_tasks)
    num_fail = random.randint(num_concurrent, num_tasks)

    tasks = [wait_task(i, random.random() / 1000, status) for i in range(num_tasks)]
    tasks += [wait_task(i, random.random() / 1000, status, fail=True) for i in range(num_fail)]
    random.shuffle(tasks)

    num_failed = 0
    i = 0

    for result in bounded_gather_iter(*tasks, limit=num_concurrent):
        try:
            result = await result
        except RuntimeError:
            num_failed += 1
            continue

        if i == quit_on:
            break

        assert 0 <= result < num_tasks
        i += 1

    assert 0 < status[1] <= num_concurrent
    assert quit_on <= status[2] <= quit_on + num_concurrent
    assert num_failed <= num_fail


def test_normalize_smartquotes():
    assert common_filters.normalize_smartquotes("Should\u2018 normalize") == "Should' normalize"
    assert common_filters.normalize_smartquotes("Same String") == "Same String"


@pytest.mark.parametrize(
    "text,pages,page_length",
    (
        # base case
        (
            "Line 1\nA longer line 2\n'tis a veeeeery long line numero tres\nand the last line",
            [
                "Line 1\nA",
                " longer line 2",
                "\n'tis a",
                " veeeeery long",
                " line numero",
                " tres\nand the",
                " last line",
            ],
            15,
        ),
        # mid-word split
        (
            "Interdisciplinary collaboration improves the quality\nof care.",
            ["Interdisciplinar", "y collaboration", " improves the", " quality\nof", " care."],
            16,
        ),
        # off-by-one errors
        ("Lorem ipsum dolor sit amet.", ["Lorem", " ipsum", " dolor", " sit", " amet."], 6),
        (
            "Lorem ipsum dolor sit amet.",
            # TODO: "r" and " sit" can fit together but current logic doesn't support it properly
            ["Lorem", " ipsu", "m", " dolo", "r", " sit", " amet", "."],
            5,
        ),
        (
            "Lorem ipsum dolor sit amet.",
            ["Lore", "m", " ips", "um", " dol", "or", " sit", " ame", "t."],
            4,
        ),
        # mass mentions
        (
            "@everyone listen to me!",
            # TODO: off-by-one: " listen" and " to me!" should have been " listen to" and " me!"
            ["@\u200beveryone", " listen", " to me!"],
            10,
        ),
        (
            "@everyone listen to me!",
            ["@everyon", "e listen", " to me!"],
            9,
        ),
        (
            "@everyone listen to me!",
            ["@everyon", "e", " listen", " to me!"],
            8,
        ),
        ("Is anyone @here?", ["Is anyone", " @\u200bhere?"], 10),
        # whitespace-only page skipping (`\n` skipped)
        ("Split:\n Long-word", ["Split:", " Long-", "word"], 6),
    ),
)
def test_pagify(text: str, pages: List[str], page_length: int):
    result = []
    for page in pagify(text, ("\n", " "), shorten_by=0, page_length=page_length):
        # sanity check
        assert len(page) <= page_length
        result.append(page)

    assert pages == result


@pytest.mark.parametrize(
    "text,pages,page_length",
    (
        # base case
        (
            "Line 1\nA longer line 2\n'tis a veeeeery long line numero tres\nand the last line",
            [
                "Line 1",
                "\nA longer line",
                " 2",
                "\n'tis a",
                " veeeeery long",
                " line numero",
                " tres",
                "\nand the last",
                " line",
            ],
            15,
        ),
        # mid-word split
        (
            "Interdisciplinary collaboration improves the quality\nof care.",
            ["Interdisciplinar", "y collaboration", " improves the", " quality", "\nof care."],
            16,
        ),
    ),
)
def test_pagify_priority(text: str, pages: List[str], page_length: int):
    result = []
    for page in pagify(text, ("\n", " "), priority=True, shorten_by=0, page_length=page_length):
        # sanity check
        assert len(page) <= page_length
        result.append(page)

    assert pages == result


def test_pagify_length_hint():
    it = pagify("A" * 100, shorten_by=0, page_length=10)
    remaining = 100 // 10

    assert operator.length_hint(it) == remaining

    for page in it:
        remaining -= 1
        assert operator.length_hint(it) == remaining

    assert operator.length_hint(it) == 0



"""
commands.converter
==================
This module contains useful functions and classes for command argument conversion.

Some of the converters within are included `provisionally <developer-guarantees-exclusions>`
and are marked as such.
"""
import functools
import math
import re
from datetime import timedelta
from dateutil.relativedelta import relativedelta
from typing import (
    TYPE_CHECKING,
    Optional,
    Optional as NoParseOptional,
    Tuple,
    List,
    Dict,
    Type,
    TypeVar,
    Union as UserInputOptional,
)

from discord.ext import commands as dpy_commands
from discord.ext.commands import BadArgument

from ..i18n import Translator
from ..utils.chat_formatting import humanize_timedelta, humanize_list

if TYPE_CHECKING:
    from .context import Context

__all__ = [
    "RawUserIdConverter",
    "DictConverter",
    "UserInputOptional",
    "NoParseOptional",
    "RelativedeltaConverter",
    "TimedeltaConverter",
    "finite_float",
    "get_dict_converter",
    "get_timedelta_converter",
    "parse_relativedelta",
    "parse_timedelta",
    "positive_int",
    "CommandConverter",
    "CogConverter",
]

_ = Translator("commands.converter", __file__)

# You'd think that Discord's documentation showing an example of 2 ** 64 - 1 snowflake would mean that
# this is going to be accepted by everything in their API but nope... Let's assume 2 ** 63 - 1 as the max instead.
ID_REGEX = re.compile(r"([0-9]{15,19})")
USER_MENTION_REGEX = re.compile(r"<@!?([0-9]{15,19})>$")
_MAX_ID = 2**63 - 1


# Taken with permission from
# https://github.com/mikeshardmind/SinbadCogs/blob/816f3bc2ba860243f75112904b82009a8a9e1f99/scheduler/time_utils.py#L9-L19
# with modifications
TIME_RE = re.compile(
    r"""
        (\s?(  # match deliminators here to make word border below unambiguous
            (?P<years>[\+-]?\d+)\s?(years?|y)
          | (?P<months>[\+-]?\d+)\s?(months?|mo)
          | (?P<weeks>[\+-]?\d+)\s?(weeks?|w)
          | (?P<days>[\+-]?\d+)\s?(days?|d)
          | (?P<hours>[\+-]?\d+)\s?(hours?|hrs|hr?)
          | (?P<minutes>[\+-]?\d+)\s?(minutes?|mins?|m)
          | (?P<seconds>[\+-]?\d+)\s?(seconds?|secs?|s)
        ))+\b
    """,
    flags=re.IGNORECASE | re.VERBOSE,
)


def _parse_and_match(string_to_match: str, allowed_units: List[str]) -> Optional[Dict[str, int]]:
    """
    Local utility function to match TIME_RE string above to user input for both parse_timedelta and parse_relativedelta
    """
    matches = TIME_RE.fullmatch(string_to_match)
    if matches:
        params = {k: int(v) for k, v in matches.groupdict().items() if v is not None}
        for k in params.keys():
            if k not in allowed_units:
                raise BadArgument(
                    _("`{unit}` is not a valid unit of time for this command").format(unit=k)
                )
        return params
    return None


def parse_timedelta(
    argument: str,
    *,
    maximum: Optional[timedelta] = None,
    minimum: Optional[timedelta] = timedelta(seconds=0),
    allowed_units: Optional[List[str]] = None,
) -> Optional[timedelta]:
    """
    This converts a user provided string into a timedelta

    If a unit is specified multiple times, only the last is considered.
    This works with or without whitespace.

    Parameters
    ----------
    argument : str
        The user provided input
    maximum : Optional[datetime.timedelta]
        If provided, any parsed value higher than this will raise an exception
    minimum : Optional[datetime.timedelta]
        If provided, any parsed value lower than this will raise an exception
        Defaults to 0 seconds, pass `datetime.timedelta.min` explicitly to allow negative values
    allowed_units : Optional[List[str]]
        If provided, you can constrain a user to expressing the amount of time
        in specific units. The units you can chose to provide are the same as the
        parser understands. (``weeks``, ``days``, ``hours``, ``minutes``, ``seconds``)

    Returns
    -------
    Optional[datetime.timedelta]
        If matched, the timedelta which was parsed. This can return `None`

    Raises
    ------
    BadArgument
        If the argument passed uses a unit not allowed, but understood
        or if the value is out of bounds.
    """
    allowed_units = allowed_units or [
        "weeks",
        "days",
        "hours",
        "minutes",
        "seconds",
    ]
    if minimum is None:
        minimum = timedelta(seconds=0)
    if maximum is None:
        maximum = timedelta.max
    params = _parse_and_match(argument, allowed_units)
    if params:
        try:
            delta = timedelta(**params)
        except OverflowError:
            raise BadArgument(
                _("The time set is way too high, consider setting something reasonable.")
            )
        if maximum < delta:
            raise BadArgument(
                _(
                    "This amount of time is too large for this command. (Maximum: {maximum})"
                ).format(
                    maximum=humanize_timedelta(seconds=math.floor(maximum.total_seconds()))
                    or _("0 seconds")
                )
            )
        if delta < minimum:
            raise BadArgument(
                _(
                    "This amount of time is too small for this command. (Minimum: {minimum})"
                ).format(
                    minimum=humanize_timedelta(seconds=math.ceil(minimum.total_seconds()))
                    or _("0 seconds")
                )
            )
        return delta
    return None


def parse_relativedelta(
    argument: str, *, allowed_units: Optional[List[str]] = None
) -> Optional[relativedelta]:
    """
    This converts a user provided string into a datetime with offset from NOW

    If a unit is specified multiple times, only the last is considered.
    This works with or without whitespace.

    Parameters
    ----------
    argument : str
        The user provided input
    allowed_units : Optional[List[str]]
        If provided, you can constrain a user to expressing the amount of time
        in specific units. The units you can chose to provide are the same as the
        parser understands. (``years``, ``months``, ``weeks``, ``days``, ``hours``, ``minutes``, ``seconds``)

    Returns
    -------
    Optional[dateutil.relativedelta.relativedelta]
        If matched, the relativedelta which was parsed. This can return `None`

    Raises
    ------
    BadArgument
        If the argument passed uses a unit not allowed, but understood
        or if the value is out of bounds.
    """
    allowed_units = allowed_units or [
        "years",
        "months",
        "weeks",
        "days",
        "hours",
        "minutes",
        "seconds",
    ]
    params = _parse_and_match(argument, allowed_units)
    if params:
        try:
            delta = relativedelta(**params)
        except OverflowError:
            raise BadArgument(
                _("The time set is way too high, consider setting something reasonable.")
            )
        return delta
    return None


class RawUserIdConverter(dpy_commands.Converter):
    """
    Converts ID or user mention to an `int`.

    Useful for commands like ``[p]ban`` or ``[p]unban`` where the bot is not necessarily
    going to share any servers with the user that a moderator wants to ban/unban.

    This converter doesn't check if the ID/mention points to an actual user
    but it won't match IDs and mentions that couldn't possibly be valid.

    For example, the converter will not match on "123" because the number doesn't have
    enough digits to be valid ID but, it will match on "12345678901234567" even though
    there is no user with such ID.
    """

    async def convert(self, ctx: "Context", argument: str) -> int:
        # This is for the hackban and unban commands, where we receive IDs that
        # are most likely not in the guild.
        # Mentions are supported, but most likely won't ever be in cache.

        if match := ID_REGEX.fullmatch(argument) or USER_MENTION_REGEX.fullmatch(argument):
            user_id = int(match.group(1))

            # Validate user ID range
            if user_id > _MAX_ID:
                raise BadArgument(
                    f"The ID '{argument}' is too large to be a valid Discord user ID."
                )

            return user_id

        raise BadArgument(_("'{input}' doesn't look like a valid user ID.").format(input=argument))


# Below this line are a lot of lies for mypy about things that *end up* correct when
# These are used for command conversion purposes. Please refer to the portion
# which is *not* for type checking for the actual implementation
# and ensure the lies stay correct for how the object should look as a typehint

#: This converts a user provided string into a positive (>=0) integer.
positive_int = dpy_commands.Range[int, 0, None]


if TYPE_CHECKING:
    finite_float = float
else:

    def finite_float(arg: str) -> float:
        """
        This converts a user provided string into a finite float.
        """
        try:
            ret = float(arg)
        except ValueError:
            raise BadArgument(_("`{arg}` is not a number.").format(arg=arg))
        if not math.isfinite(ret):
            raise BadArgument(_("`{arg}` is not a finite number.").format(arg=ret))
        return ret


if TYPE_CHECKING:
    DictConverter = Dict[str, str]
else:

    class DictConverter(dpy_commands.Converter):
        """
        Converts pairs of space separated values to a dict
        """

        def __init__(self, *expected_keys: str, delims: Optional[List[str]] = None):
            self.expected_keys = expected_keys
            self.delims = delims or [" "]
            self.pattern = re.compile(r"|".join(re.escape(d) for d in self.delims))

        async def convert(self, ctx: "Context", argument: str) -> Dict[str, str]:
            ret: Dict[str, str] = {}
            args = self.pattern.split(argument)

            if len(args) % 2 != 0:
                raise BadArgument(_("Missing a key or value."))

            iterator = iter(args)

            for key in iterator:
                if self.expected_keys and key not in self.expected_keys:
                    raise BadArgument(_("Unexpected key `{key}`.").format(key=key))

                ret[key] = next(iterator)

            return ret


if TYPE_CHECKING:

    def get_dict_converter(*expected_keys: str, delims: Optional[List[str]] = None) -> Type[dict]:
        ...

else:

    def get_dict_converter(*expected_keys: str, delims: Optional[List[str]] = None) -> Type[dict]:
        """
        Returns a typechecking safe `DictConverter` suitable for use with discord.py
        """

        class PartialMeta(type(DictConverter)):
            __call__ = functools.partialmethod(
                type(DictConverter).__call__, *expected_keys, delims=delims
            )

        class ValidatedConverter(DictConverter, metaclass=PartialMeta):
            pass

        return ValidatedConverter


if TYPE_CHECKING:
    TimedeltaConverter = timedelta
else:

    class TimedeltaConverter(dpy_commands.Converter):
        """
        This is a converter for timedeltas.
        The units should be in order from largest to smallest.
        This works with or without whitespace.

        See `parse_timedelta` for more information about how this functions.

        Attributes
        ----------
        maximum : Optional[datetime.timedelta]
            If provided, any parsed value higher than this will raise an exception
        minimum : Optional[datetime.timedelta]
            If provided, any parsed value lower than this will raise an exception
            Defaults to 0 seconds, pass `datetime.timedelta.min` explicitly to allow negative values
        allowed_units : Optional[List[str]]
            If provided, you can constrain a user to expressing the amount of time
            in specific units. The units you can choose to provide are the same as the
            parser understands: (``weeks``, ``days``, ``hours``, ``minutes``, ``seconds``)
        default_unit : Optional[str]
            If provided, it will additionally try to match integer-only input into
            a timedelta, using the unit specified. Same units as in ``allowed_units``
            apply.
        """

        def __init__(
            self,
            *,
            minimum=timedelta(seconds=0),
            maximum=None,
            allowed_units=None,
            default_unit=None,
        ):
            self.allowed_units = allowed_units
            self.default_unit = default_unit
            self.minimum = minimum
            self.maximum = maximum

        async def convert(self, ctx: "Context", argument: str) -> timedelta:
            if self.default_unit and argument.isdecimal():
                argument = argument + self.default_unit

            delta = parse_timedelta(
                argument,
                minimum=self.minimum,
                maximum=self.maximum,
                allowed_units=self.allowed_units,
            )

            if delta is not None:
                return delta
            raise BadArgument()  # This allows this to be a required argument.


if TYPE_CHECKING:

    def get_timedelta_converter(
        *,
        default_unit: Optional[str] = None,
        maximum: Optional[timedelta] = None,
        minimum: Optional[timedelta] = timedelta(seconds=0),
        allowed_units: Optional[List[str]] = None,
    ) -> Type[timedelta]:
        ...

else:

    def get_timedelta_converter(
        *,
        default_unit: Optional[str] = None,
        maximum: Optional[timedelta] = None,
        minimum: Optional[timedelta] = timedelta(seconds=0),
        allowed_units: Optional[List[str]] = None,
    ) -> Type[timedelta]:
        """
        This creates a type suitable for typechecking which works with discord.py's
        commands.

        See `parse_timedelta` for more information about how this functions.

        Parameters
        ----------
        maximum : Optional[datetime.timedelta]
            If provided, any parsed value higher than this will raise an exception
        minimum : Optional[datetime.timedelta]
            If provided, any parsed value lower than this will raise an exception
            Defaults to 0 seconds, pass `datetime.timedelta.min` explicitly to allow negative values
        allowed_units : Optional[List[str]]
            If provided, you can constrain a user to expressing the amount of time
            in specific units. The units you can choose to provide are the same as the
            parser understands: (``weeks``, ``days``, ``hours``, ``minutes``, ``seconds``)
        default_unit : Optional[str]
            If provided, it will additionally try to match integer-only input into
            a timedelta, using the unit specified. Same units as in ``allowed_units``
            apply.

        Returns
        -------
        type
            The converter class, which will be a subclass of `TimedeltaConverter`
        """

        class PartialMeta(type(DictConverter)):
            __call__ = functools.partialmethod(
                type(DictConverter).__call__,
                allowed_units=allowed_units,
                default_unit=default_unit,
                minimum=minimum,
                maximum=maximum,
            )

        class ValidatedConverter(TimedeltaConverter, metaclass=PartialMeta):
            pass

        return ValidatedConverter


if TYPE_CHECKING:
    RelativedeltaConverter = relativedelta
else:

    class RelativedeltaConverter(dpy_commands.Converter):
        """
        This is a converter for relative deltas.

        The units should be in order from largest to smallest.
        This works with or without whitespace.

        See `parse_relativedelta` for more information about how this functions.

        Attributes
        ----------
        allowed_units : Optional[List[str]]
            If provided, you can constrain a user to expressing the amount of time
            in specific units. The units you can choose to provide are the same as the
            parser understands: (``years``, ``months``, ``weeks``, ``days``, ``hours``, ``minutes``, ``seconds``)
        default_unit : Optional[str]
            If provided, it will additionally try to match integer-only input into
            a timedelta, using the unit specified. Same units as in ``allowed_units``
            apply.
        """

        def __init__(self, *, allowed_units=None, default_unit=None):
            self.allowed_units = allowed_units
            self.default_unit = default_unit

        async def convert(self, ctx: "Context", argument: str) -> relativedelta:
            if self.default_unit and argument.isdecimal():
                argument = argument + self.default_unit

            delta = parse_relativedelta(argument, allowed_units=self.allowed_units)

            if delta is not None:
                return delta
            raise BadArgument()  # This allows this to be a required argument.


if not TYPE_CHECKING:

    class NoParseOptional:
        """
        This can be used instead of `typing.Optional`
        to avoid discord.py special casing the conversion behavior.

        .. seealso::
            The `ignore_optional_for_conversion` option of commands.
        """

        def __class_getitem__(cls, key):
            if isinstance(key, tuple):
                raise TypeError("Must only provide a single type to Optional")
            return key


_T = TypeVar("_T")

if not TYPE_CHECKING:
    #: This can be used when user input should be converted as discord.py
    #: treats `typing.Optional`, but the type should not be equivalent to
    #: ``typing.Union[DesiredType, None]`` for type checking.
    #:
    #: Note: In type checking context, this type hint can be passed
    #: multiple types, but such usage is not supported and will fail at runtime
    #:
    #: .. warning::
    #:    This converter class is still `provisional <developer-guarantees-exclusions>`.
    UserInputOptional = Optional

if TYPE_CHECKING:
    CommandConverter = dpy_commands.Command
    CogConverter = dpy_commands.Cog
else:

    class CommandConverter(dpy_commands.Converter):
        """Converts a command name to the matching `redbot.core.commands.Command` object."""

        async def convert(self, ctx: "Context", argument: str):
            arg = argument.strip()
            command = ctx.bot.get_command(arg)
            if not command:
                raise BadArgument(_('Command "{arg}" not found.').format(arg=arg))
            return command

    class CogConverter(dpy_commands.Converter):
        """Converts a cog name to the matching `redbot.core.commands.Cog` object."""

        async def convert(self, ctx: "Context", argument: str):
            arg = argument.strip()
            cog = ctx.bot.get_cog(arg)
            if not cog:
                raise BadArgument(_('Cog "{arg}" not found.').format(arg=arg))
            return cog



from datetime import datetime, timedelta
from typing import Tuple, List
from collections import namedtuple

__all__ = ("AntiSpam",)

_AntiSpamInterval = namedtuple("_AntiSpamInterval", ["period", "frequency"])


class AntiSpam:
    """
    A class that can be used to count the number of events that happened
    within specified intervals. Later, it can be checked whether the specified
    maximum count for any of the specified intervals has been exceeded.

    Examples
    --------
    Tracking whether the number of reports sent by a user within a single guild is spammy:

    .. code-block:: python

        class MyCog(commands.Cog):
            INTERVALS = [
                # More than one report within last 5 seconds is considered spam.
                (datetime.timedelta(seconds=5), 1),
                # More than 3 reports within the last 5 minutes are considered spam.
                (datetime.timedelta(minutes=5), 3),
                # More than 10 reports within the last hour are considered spam.
                (datetime.timedelta(hours=1), 10),
                # More than 24 reports within a single day (last 24 hours) are considered spam.
                (datetime.timedelta(days=1), 24),
            ]

            def __init__(self, bot):
                self.bot = bot
                self.antispam = {}

            @commands.guild_only()
            @commands.command()
            async def report(self, ctx, content):
                # We want to track whether a single user within a single guild
                # sends a spammy number of reports.
                key = (ctx.guild.id, ctx.author.id)

                if key not in self.antispam:
                    # Create an instance of the AntiSpam class with given intervals.
                    self.antispam[key] = AntiSpam(self.INTERVALS)
                    # If you want to use the default intervals, you can use: AntiSpam([])

                # Check if the user sent too many reports recently.
                # The `AntiSpam.spammy` property is `True` if, for any interval,
                # the number of events that happened within that interval
                # exceeds the number specified for that interval.
                if self.antispam[key].spammy:
                    await ctx.send(
                        "You've sent too many reports recently, please try again later."
                    )
                    return

                # Make any other early-return checks.

                # Record the event.
                self.antispam[key].stamp()

                # Save the report.
                # self.config...

                # Send a message to the user.
                await ctx.send("Your report has been submitted.")

    Parameters
    ----------
    intervals : List[Tuple[datetime.timedelta, int]]
        A list of tuples in the format (timedelta, int),
        where the timedelta represents the length of the interval,
        and the int represents the maximum number of times something
        can happen within that interval.

        If an empty list is provided, the following defaults will be used:

        * 3 per 5 seconds
        * 5 per 1 minute
        * 10 per 1 hour
        * 24 per 1 day
    """

    # TODO : Decorator interface for command check using `spammy`
    # with insertion of the antispam element into context
    # for manual stamping on successful command completion

    default_intervals = [
        (timedelta(seconds=5), 3),
        (timedelta(minutes=1), 5),
        (timedelta(hours=1), 10),
        (timedelta(days=1), 24),
    ]

    def __init__(self, intervals: List[Tuple[timedelta, int]]):
        self.__event_timestamps = []
        _itvs = intervals or self.default_intervals
        self.__intervals = [_AntiSpamInterval(*x) for x in _itvs]
        self.__discard_after = max([x.period for x in self.__intervals])

    def __interval_check(self, interval: _AntiSpamInterval):
        return (
            len([t for t in self.__event_timestamps if (t + interval.period) > datetime.utcnow()])
            >= interval.frequency
        )

    @property
    def spammy(self):
        """
        Whether, for any interval, the number of events that happened
        within that interval exceeds the number specified for that interval.
        """
        return any(self.__interval_check(x) for x in self.__intervals)

    def stamp(self):
        """
        Mark an event timestamp against the list of antispam intervals happening right now.
        Use this after all checks have passed, and your action is taking place.

        The stamp will last until the corresponding interval duration
        has expired (set when this AntiSpam object was initiated).
        """
        self.__event_timestamps.append(datetime.utcnow())
        self.__event_timestamps = [
            t for t in self.__event_timestamps if t + self.__discard_after > datetime.utcnow()
        ]



from __future__ import annotations

import logging
import re
from typing import Optional, TypedDict
from datetime import timedelta, datetime, timezone
from typing_extensions import Annotated

from discord.ext.commands.converter import Converter
from redbot.core import commands
from redbot.core import i18n
from redbot.core.commands.converter import TIME_RE

_ = i18n.Translator("Mutes", __file__)
log = logging.getLogger("red.cogs.mutes")

TIME_SPLIT = re.compile(r"t(?:ime\s?)?=\s*")


def _edgematch(pattern: re.Pattern[str], argument: str) -> Optional[re.Match[str]]:
    """Internal utility to match at either end of the argument string"""
    # precondition: pattern does not end in $
    # precondition: argument does not end in whitespace
    return pattern.match(argument) or re.search(
        pattern.pattern + "$", argument, flags=pattern.flags
    )


class _MuteTime(TypedDict, total=False):
    duration: timedelta
    reason: str
    until: datetime


class _MuteTimeConverter(Converter):
    """
    This will parse my defined multi response pattern and provide usable formats
    to be used in multiple responses
    """

    async def convert(self, ctx: commands.Context, argument: str) -> _MuteTime:
        time_split = TIME_SPLIT.search(argument)
        result: _MuteTime = {}
        if time_split:
            maybe_time = argument[time_split.end() :]
            strategy = re.match
        else:
            maybe_time = argument
            strategy = _edgematch

        match = strategy(TIME_RE, maybe_time)
        if match:
            time_data = {k: int(v) for k, v in match.groupdict().items() if v is not None}
            for k in time_data:
                if k in ("years", "months"):
                    raise commands.BadArgument(
                        _("`{unit}` is not a valid unit of time for this command").format(unit=k)
                    )
            try:
                result["duration"] = duration = timedelta(**time_data)
                result["until"] = ctx.message.created_at + duration
                # Catch if using the timedelta with the current date will also result in an Overflow error
            except OverflowError:
                raise commands.BadArgument(
                    _("The time provided is too long; use a more reasonable time.")
                )
            if duration <= timedelta(seconds=0):
                raise commands.BadArgument(_("The time provided must not be in the past."))
            if time_split:
                start, end = time_split.span()
                end += match.end()
            else:
                start, end = match.span()
            argument = argument[:start] + argument[end:]
        result["reason"] = argument.strip()
        return result


MuteTime = Annotated[_MuteTime, _MuteTimeConverter]



from __future__ import annotations

from contextvars import ContextVar
from typing import TYPE_CHECKING, List, Optional

from babel.core import Locale, UnknownLocaleError

if TYPE_CHECKING:
    from redbot.core.i18n import Translator


__all__ = (
    "current_locale",
    "current_locale_default",
    "current_regional_format",
    "current_regional_format_default",
    "translators",
    "set_global_locale",
    "set_global_regional_format",
    "set_contextual_locale",
    "set_contextual_regional_format",
)


FRESH_INSTALL_LOCALE = "en-US"
current_locale = ContextVar("current_locale")
current_locale_default = FRESH_INSTALL_LOCALE
current_regional_format = ContextVar("current_regional_format")
current_regional_format_default = None

translators: List[Translator] = []


def _reload_locales() -> None:
    for translator in translators:
        translator.load_translations()


def _get_standardized_locale_name(language_code: str) -> str:
    try:
        locale = Locale.parse(language_code, sep="-")
    except (ValueError, UnknownLocaleError):
        raise ValueError("Invalid language code. Use format: `en-US`")
    if locale.territory is None:
        raise ValueError(
            "Invalid format - language code has to include country code, e.g. `en-US`"
        )
    return f"{locale.language}-{locale.territory}"


def set_global_locale(language_code: str, /) -> str:
    global current_locale_default
    current_locale_default = _get_standardized_locale_name(language_code)
    _reload_locales()
    return current_locale_default


def set_global_regional_format(language_code: Optional[str], /) -> Optional[str]:
    global current_regional_format_default
    if language_code is not None:
        language_code = _get_standardized_locale_name(language_code)
    current_regional_format_default = language_code
    return language_code


def set_contextual_locale(language_code: str, /, verify_language_code: bool = False) -> str:
    if verify_language_code:
        language_code = _get_standardized_locale_name(language_code)
    current_locale.set(language_code)
    _reload_locales()
    return language_code


def set_contextual_regional_format(
    language_code: str, /, verify_language_code: bool = False
) -> str:
    if verify_language_code and language_code is not None:
        language_code = _get_standardized_locale_name(language_code)
    current_regional_format.set(language_code)
    return language_code



from typing import Tuple, Dict, Optional, List, Union
from re import findall

import discord
from discord.ext.commands.view import StringView  # DEP-WARN
from redbot.core import commands, Config
from redbot.core.i18n import Translator
from redbot.core.utils import AsyncIter

_ = Translator("Alias", __file__)


class ArgParseError(Exception):
    pass


class AliasEntry:
    """An object containing all required information about an alias"""

    name: str
    command: Union[Tuple[str], str]
    creator: int
    guild: Optional[int]
    uses: int

    def __init__(
        self, name: str, command: Union[Tuple[str], str], creator: int, guild: Optional[int]
    ):
        super().__init__()
        self.name = name
        self.command = command
        self.creator = creator

        self.guild = guild
        self.uses = 0

    def inc(self):
        """
        Increases the `uses` stat by 1.
        :return: new use count
        """
        self.uses += 1
        return self.uses

    def get_extra_args_from_alias(self, message: discord.Message, prefix: str) -> str:
        """
        When an alias is executed by a user in chat this function tries
            to get any extra arguments passed in with the call.
            Whitespace will be trimmed from both ends.
        :param message:
        :param prefix:
        :param alias:
        :return:
        """
        known_content_length = len(prefix) + len(self.name)
        extra = message.content[known_content_length:]
        view = StringView(extra)
        view.skip_ws()
        extra = []
        while not view.eof:
            prev = view.index
            word = view.get_quoted_word()
            if len(word) < view.index - prev:
                word = "".join((view.buffer[prev], word, view.buffer[view.index - 1]))
            extra.append(word.strip(" "))
        return extra

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "command": self.command,
            "creator": self.creator,
            "guild": self.guild,
            "uses": self.uses,
        }

    @classmethod
    def from_json(cls, data: dict):
        ret = cls(data["name"], data["command"], data["creator"], data["guild"])
        ret.uses = data.get("uses", 0)
        return ret


class AliasCache:
    def __init__(self, config: Config, cache_enabled: bool = True):
        self.config = config
        self._cache_enabled = cache_enabled
        self._loaded = False
        self._aliases: Dict[Optional[int], Dict[str, AliasEntry]] = {None: {}}

    async def anonymize_aliases(self, user_id: int):
        async with self.config.entries() as global_aliases:
            for a in global_aliases:
                if a.get("creator", 0) == user_id:
                    a["creator"] = 0xDE1
                    if self._cache_enabled:
                        self._aliases[None][a["name"]] = AliasEntry.from_json(a)

        all_guilds = await self.config.all_guilds()
        async for guild_id, guild_data in AsyncIter(all_guilds.items(), steps=100):
            for a in guild_data["entries"]:
                if a.get("creator", 0) == user_id:
                    break
            else:
                continue
            # basically, don't build a context manager without a need.
            async with self.config.guild_from_id(guild_id).entries() as entry_list:
                for a in entry_list:
                    if a.get("creator", 0) == user_id:
                        a["creator"] = 0xDE1
                        if self._cache_enabled:
                            self._aliases[guild_id][a["name"]] = AliasEntry.from_json(a)

    async def load_aliases(self):
        if not self._cache_enabled:
            self._loaded = True
            return
        for alias in await self.config.entries():
            self._aliases[None][alias["name"]] = AliasEntry.from_json(alias)

        all_guilds = await self.config.all_guilds()
        async for guild_id, guild_data in AsyncIter(all_guilds.items(), steps=100):
            if guild_id not in self._aliases:
                self._aliases[guild_id] = {}
            for alias in guild_data["entries"]:
                self._aliases[guild_id][alias["name"]] = AliasEntry.from_json(alias)
        self._loaded = True

    async def get_aliases(self, ctx: commands.Context) -> List[AliasEntry]:
        """Returns all possible aliases with the given context"""
        global_aliases: List[AliasEntry] = []
        server_aliases: List[AliasEntry] = []
        global_aliases = await self.get_global_aliases()
        if ctx.guild and ctx.guild.id in self._aliases:
            server_aliases = await self.get_guild_aliases(ctx.guild)
        return global_aliases + server_aliases

    async def get_guild_aliases(self, guild: discord.Guild) -> List[AliasEntry]:
        """Returns all guild specific aliases"""
        aliases: List[AliasEntry] = []

        if self._cache_enabled:
            if guild.id in self._aliases:
                for _, alias in self._aliases[guild.id].items():
                    aliases.append(alias)
        else:
            aliases = [AliasEntry.from_json(d) for d in await self.config.guild(guild).entries()]
        return aliases

    async def get_global_aliases(self) -> List[AliasEntry]:
        """Returns all global specific aliases"""
        aliases: List[AliasEntry] = []
        if self._cache_enabled:
            for _, alias in self._aliases[None].items():
                aliases.append(alias)
        else:
            aliases = [AliasEntry.from_json(d) for d in await self.config.entries()]
        return aliases

    async def get_alias(
        self, guild: Optional[discord.Guild], alias_name: str
    ) -> Optional[AliasEntry]:
        """Returns an AliasEntry object if the provided alias_name is a registered alias"""
        server_aliases: List[AliasEntry] = []

        if self._cache_enabled:
            if alias_name in self._aliases[None]:
                return self._aliases[None][alias_name]
            if guild is not None:
                if guild.id in self._aliases:
                    if alias_name in self._aliases[guild.id]:
                        return self._aliases[guild.id][alias_name]
        else:
            if guild:
                server_aliases = [
                    AliasEntry.from_json(d) for d in await self.config.guild(guild.id).entries()
                ]
            global_aliases = [AliasEntry.from_json(d) for d in await self.config.entries()]
            all_aliases = global_aliases + server_aliases

            for alias in all_aliases:
                if alias.name == alias_name:
                    return alias

        return None

    @staticmethod
    def format_command_for_alias(command: str) -> str:
        # This was present in add_alias previously
        # Made this into a separate method so as to reuse the same code in edit_alias
        indices = findall(r"{(\d*)}", command)
        if indices:
            try:
                indices = [int(a[0]) for a in indices]
            except IndexError:
                raise ArgParseError(_("Arguments must be specified with a number."))
            low = min(indices)
            indices = [a - low for a in indices]
            high = max(indices)
            gaps = set(indices).symmetric_difference(range(high + 1))
            if gaps:
                raise ArgParseError(
                    _("Arguments must be sequential. Missing arguments: ")
                    + ", ".join(str(i + low) for i in gaps)
                )
            command = command.format(*(f"{{{i}}}" for i in range(-low, high + low + 1)))
        return command

    async def add_alias(
        self, ctx: commands.Context, alias_name: str, command: str, global_: bool = False
    ) -> AliasEntry:
        command = self.format_command_for_alias(command)

        if global_:
            alias = AliasEntry(alias_name, command, ctx.author.id, None)
            settings = self.config
            if self._cache_enabled:
                self._aliases[None][alias.name] = alias
        else:
            alias = AliasEntry(alias_name, command, ctx.author.id, ctx.guild.id)
            settings = self.config.guild(ctx.guild)
            if self._cache_enabled:
                if ctx.guild.id not in self._aliases:
                    self._aliases[ctx.guild.id] = {}
                self._aliases[ctx.guild.id][alias.name] = alias

        async with settings.entries() as curr_aliases:
            curr_aliases.append(alias.to_json())

        return alias

    async def edit_alias(
        self, ctx: commands.Context, alias_name: str, command: str, global_: bool = False
    ) -> bool:
        command = self.format_command_for_alias(command)

        if global_:
            settings = self.config
        else:
            settings = self.config.guild(ctx.guild)

        async with settings.entries() as aliases:
            for index, alias in enumerate(aliases):
                if alias["name"] == alias_name:
                    alias_edited = AliasEntry.from_json(alias)
                    alias_edited.command = command
                    aliases[index] = alias_edited.to_json()

                    if self._cache_enabled:
                        if global_:
                            self._aliases[None][alias_edited.name] = alias_edited
                        else:
                            self._aliases[ctx.guild.id][alias_edited.name] = alias_edited
                    return True

        return False

    async def delete_alias(
        self, ctx: commands.Context, alias_name: str, global_: bool = False
    ) -> bool:
        if global_:
            settings = self.config
        else:
            settings = self.config.guild(ctx.guild)

        async with settings.entries() as aliases:
            for alias in aliases:
                if alias["name"] == alias_name:
                    aliases.remove(alias)
                    if self._cache_enabled:
                        if global_:
                            del self._aliases[None][alias_name]
                        else:
                            del self._aliases[ctx.guild.id][alias_name]
                    return True

        return False



import math
from pathlib import Path

from typing import List, Tuple

import discord
import lavalink
from red_commons.logging import getLogger

import rapidfuzz
from redbot.core import commands
from redbot.core.i18n import Translator
from redbot.core.utils import AsyncIter
from redbot.core.utils.chat_formatting import humanize_number

from ...audio_dataclasses import LocalPath, Query
from ..abc import MixinMeta
from ..cog_utils import CompositeMetaClass

log = getLogger("red.cogs.Audio.cog.Utilities.queue")
_ = Translator("Audio", Path(__file__))


class QueueUtilities(MixinMeta, metaclass=CompositeMetaClass):
    async def _build_queue_page(
        self,
        ctx: commands.Context,
        queue: list,
        player: lavalink.player.Player,
        page_num: int,
    ) -> discord.Embed:
        shuffle = await self.config.guild(ctx.guild).shuffle()
        repeat = await self.config.guild(ctx.guild).repeat()
        autoplay = await self.config.guild(ctx.guild).auto_play()

        queue_num_pages = math.ceil(len(queue) / 10)
        queue_idx_start = (page_num - 1) * 10
        queue_idx_end = queue_idx_start + 10
        if len(player.queue) > 500:
            queue_list = _("__Too many songs in the queue, only showing the first 500__.\n\n")
        else:
            queue_list = ""

        arrow = await self.draw_time(ctx)
        pos = self.format_time(player.position)

        if player.current.is_stream:
            dur = "LIVE"
        else:
            dur = self.format_time(player.current.length)

        query = Query.process_input(player.current, self.local_folder_current_path)
        current_track_description = await self.get_track_description(
            player.current, self.local_folder_current_path
        )
        if query.is_stream:
            queue_list += _("**Currently livestreaming:**\n")
            queue_list += f"{current_track_description}\n"
            queue_list += _("Requested by: **{user}**").format(user=player.current.requester)
            queue_list += f"\n\n{arrow}`{pos}`/`{dur}`\n\n"
        else:
            queue_list += _("Playing: ")
            queue_list += f"{current_track_description}\n"
            queue_list += _("Requested by: **{user}**").format(user=player.current.requester)
            queue_list += f"\n\n{arrow}`{pos}`/`{dur}`\n\n"

        async for i, track in AsyncIter(queue[queue_idx_start:queue_idx_end]).enumerate(
            start=queue_idx_start
        ):
            req_user = track.requester
            track_idx = i + 1
            track_description = await self.get_track_description(
                track, self.local_folder_current_path, shorten=True
            )
            queue_list += f"`{track_idx}.` {track_description}, "
            queue_list += _("requested by **{user}**\n").format(user=req_user)

        embed = discord.Embed(
            colour=await ctx.embed_colour(),
            title=_("Queue for __{guild_name}__").format(guild_name=ctx.guild.name),
            description=queue_list,
        )

        if await self.config.guild(ctx.guild).thumbnail() and player.current.thumbnail:
            embed.set_thumbnail(url=player.current.thumbnail)
        queue_dur = await self.queue_duration(ctx)
        queue_total_duration = self.format_time(queue_dur)
        text = _(
            "Page {page_num}/{total_pages} | {num_tracks} tracks, {num_remaining} remaining\n"
        ).format(
            page_num=humanize_number(page_num),
            total_pages=humanize_number(queue_num_pages),
            num_tracks=len(player.queue),
            num_remaining=queue_total_duration,
        )
        text += (
            _("Auto-Play")
            + ": "
            + ("\N{WHITE HEAVY CHECK MARK}" if autoplay else "\N{CROSS MARK}")
        )
        text += (
            (" | " if text else "")
            + _("Shuffle")
            + ": "
            + ("\N{WHITE HEAVY CHECK MARK}" if shuffle else "\N{CROSS MARK}")
        )
        text += (
            (" | " if text else "")
            + _("Repeat")
            + ": "
            + ("\N{WHITE HEAVY CHECK MARK}" if repeat else "\N{CROSS MARK}")
        )
        embed.set_footer(text=text)
        return embed

    async def _build_queue_search_list(
        self, queue_list: List[lavalink.Track], search_words: str
    ) -> List[Tuple[int, str]]:
        tracks = {}
        async for queue_idx, track in AsyncIter(queue_list).enumerate(start=1):
            if not self.match_url(track.uri):
                query = Query.process_input(track, self.local_folder_current_path)
                if (
                    query.is_local
                    and query.local_track_path is not None
                    and track.title == "Unknown title"
                ):
                    track_title = query.local_track_path.to_string_user()
                else:
                    track_title = "{} - {}".format(track.author, track.title)
            else:
                track_title = track.title

            tracks[queue_idx] = track_title
        search_results = rapidfuzz.process.extract(
            search_words, tracks, limit=50, processor=rapidfuzz.utils.default_process
        )
        search_list = []
        async for title, percent_match, queue_position in AsyncIter(search_results):
            if percent_match > 89:
                search_list.append((queue_position, title))
        return search_list

    async def _build_queue_search_page(
        self, ctx: commands.Context, page_num: int, search_list: List[Tuple[int, str]]
    ) -> discord.Embed:
        search_num_pages = math.ceil(len(search_list) / 10)
        search_idx_start = (page_num - 1) * 10
        search_idx_end = search_idx_start + 10
        track_match = ""
        async for i, track in AsyncIter(search_list[search_idx_start:search_idx_end]).enumerate(
            start=search_idx_start
        ):
            track_idx = i + 1
            if type(track) is str:
                track_location = LocalPath(track, self.local_folder_current_path).to_string_user()
                track_match += "`{}.` **{}**\n".format(track_idx, track_location)
            else:
                track_match += "`{}.` **{}**\n".format(track[0], track[1])
        embed = discord.Embed(
            colour=await ctx.embed_colour(), title=_("Matching Tracks:"), description=track_match
        )
        embed.set_footer(
            text=_("Page {page_num}/{total_pages} | {num_tracks} tracks").format(
                page_num=humanize_number(page_num),
                total_pages=humanize_number(search_num_pages),
                num_tracks=len(search_list),
            )
        )
        return embed



"""Package for Trivia cog."""
from redbot.core.bot import Red

from .trivia import *
from .session import *
from .log import *


async def setup(bot: Red) -> None:
    """Load Trivia."""
    await bot.add_cog(Trivia(bot))



import asyncio

from redbot.core import commands
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import bold
from redbot.core.utils.predicates import MessagePredicate

__all__ = ["do_install_agreement"]

_ = Translator("DownloaderChecks", __file__)


async def do_install_agreement(ctx: commands.Context) -> bool:
    downloader = ctx.cog
    if downloader is None or downloader.already_agreed:
        return True

    confirmation_message = "I agree"
    await ctx.send(
        _(
            "You're about to add a 3rd party repository. The creator of Red"
            " and its community have no responsibility for any potential "
            "damage that the content of 3rd party repositories might cause."
            "\n\nBy typing '{confirmation_message}' you declare that you have read and"
            " fully understand the above message. This message won't be "
            "shown again until the next reboot.\n\nYou have **30** seconds"
            " to reply to this message."
        ).format(confirmation_message=bold(confirmation_message))
    )

    try:
        await ctx.bot.wait_for(
            "message",
            check=MessagePredicate.lower_equal_to(confirmation_message.lower(), ctx),
            timeout=30,
        )
    except asyncio.TimeoutError:
        await ctx.send(_("Your response has timed out, please try again."))
        return False

    downloader.already_agreed = True
    return True



from pathlib import Path

from typing import List, MutableMapping, Optional, Union

import discord
import lavalink
from red_commons.logging import getLogger

from redbot.core import Config, commands
from redbot.core.bot import Red
from redbot.core.i18n import Translator
from redbot.core.utils import AsyncIter

from ..errors import NotAllowed
from ..utils import PlaylistScope
from .api_utils import PlaylistFetchResult, prepare_config_scope, standardize_scope
from .playlist_wrapper import PlaylistWrapper

log = getLogger("red.cogs.Audio.api.PlaylistsInterface")
_ = Translator("Audio", Path(__file__))


class Playlist:
    """A single playlist."""

    def __init__(
        self,
        bot: Red,
        playlist_api: PlaylistWrapper,
        scope: str,
        author: int,
        playlist_id: int,
        name: str,
        playlist_url: Optional[str] = None,
        tracks: Optional[List[MutableMapping]] = None,
        guild: Union[discord.Guild, int, None] = None,
    ):
        self.bot = bot
        self.guild = guild
        self.scope = standardize_scope(scope)
        self.config_scope = prepare_config_scope(self.bot, self.scope, author, guild)
        self.scope_id = self.config_scope[-1]
        self.author = author
        self.author_id = getattr(self.author, "id", self.author)
        self.guild_id = (
            getattr(guild, "id", guild) if self.scope == PlaylistScope.GLOBAL.value else None
        )
        self.id = playlist_id
        self.name = name
        self.url = playlist_url
        self.tracks = tracks or []
        self.tracks_obj = [lavalink.Track(data=track) for track in self.tracks]
        self.playlist_api = playlist_api

    def __repr__(self):
        return (
            f"Playlist(name={self.name}, id={self.id}, scope={self.scope}, "
            f"scope_id={self.scope_id}, author={self.author_id}, "
            f"tracks={len(self.tracks)}, url={self.url})"
        )

    async def edit(self, data: MutableMapping):
        """
        Edits a Playlist.
        Parameters
        ----------
        data: dict
            The attributes to change.
        """
        # Disallow ID editing
        if "id" in data:
            raise NotAllowed("Playlist ID cannot be edited.")

        for item in list(data.keys()):
            setattr(self, item, data[item])
        await self.save()
        return self

    async def save(self):
        """Saves a Playlist."""
        scope, scope_id = self.config_scope
        await self.playlist_api.upsert(
            scope,
            playlist_id=int(self.id),
            playlist_name=self.name,
            scope_id=scope_id,
            author_id=self.author_id,
            playlist_url=self.url,
            tracks=self.tracks,
        )

    def to_json(self) -> MutableMapping:
        """Transform the object to a dict.
        Returns
        -------
        dict
            The playlist in the form of a dict.
        """
        data = dict(
            id=self.id,
            author=self.author_id,
            guild=self.guild_id,
            name=self.name,
            playlist_url=self.url,
            tracks=self.tracks,
        )

        return data

    @classmethod
    async def from_json(
        cls,
        bot: Red,
        playlist_api: PlaylistWrapper,
        scope: str,
        playlist_number: int,
        data: PlaylistFetchResult,
        **kwargs,
    ) -> "Playlist":
        """Get a Playlist object from the provided information.
        Parameters
        ----------
        bot: Red
            The bot's instance. Needed to get the target user.
        playlist_api: PlaylistWrapper
            The Playlist API interface.
        scope:str
            The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
        playlist_number: int
            The playlist's number.
        data: PlaylistFetchResult
            The PlaylistFetchResult representation of the playlist to be gotten.
        **kwargs
            Extra attributes for the Playlist instance which override values
            in the data dict. These should be complete objects and not
            IDs, where possible.
        Returns
        -------
        Playlist
            The playlist object for the requested playlist.
        Raises
        ------
        `InvalidPlaylistScope`
            Passing a scope that is not supported.
        `MissingGuild`
            Trying to access the Guild scope without a guild.
        `MissingAuthor`
            Trying to access the User scope without an user id.
        """
        guild = data.scope_id if scope == PlaylistScope.GUILD.value else kwargs.get("guild")
        author = data.author_id
        playlist_id = data.playlist_id or playlist_number
        name = data.playlist_name
        playlist_url = data.playlist_url
        tracks = data.tracks

        return cls(
            bot=bot,
            playlist_api=playlist_api,
            guild=guild,
            scope=scope,
            author=author,
            playlist_id=playlist_id,
            name=name,
            playlist_url=playlist_url,
            tracks=tracks,
        )


class PlaylistCompat23:
    """A single playlist, migrating from Schema 2 to Schema 3"""

    def __init__(
        self,
        bot: Red,
        playlist_api: PlaylistWrapper,
        scope: str,
        author: int,
        playlist_id: int,
        name: str,
        playlist_url: Optional[str] = None,
        tracks: Optional[List[MutableMapping]] = None,
        guild: Union[discord.Guild, int, None] = None,
    ):
        self.bot = bot
        self.guild = guild
        self.scope = standardize_scope(scope)
        self.author = author
        self.id = playlist_id
        self.name = name
        self.url = playlist_url
        self.tracks = tracks or []

        self.playlist_api = playlist_api

    @classmethod
    async def from_json(
        cls,
        bot: Red,
        playlist_api: PlaylistWrapper,
        scope: str,
        playlist_number: int,
        data: MutableMapping,
        **kwargs,
    ) -> "PlaylistCompat23":
        """Get a Playlist object from the provided information.
        Parameters
        ----------
        bot: Red
            The Bot instance.
        playlist_api: PlaylistWrapper
            The Playlist API interface.
        scope:str
            The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
        playlist_number: int
            The playlist's number.
        data: MutableMapping
            The JSON representation of the playlist to be gotten.
        **kwargs
            Extra attributes for the Playlist instance which override values
            in the data dict. These should be complete objects and not
            IDs, where possible.
        Returns
        -------
        Playlist
            The playlist object for the requested playlist.
        Raises
        ------
        `InvalidPlaylistScope`
            Passing a scope that is not supported.
        `MissingGuild`
            Trying to access the Guild scope without a guild.
        `MissingAuthor`
            Trying to access the User scope without an user id.
        """
        guild = data.get("guild") or kwargs.get("guild")
        author: int = data.get("author") or kwargs.get("author") or 0
        playlist_id = data.get("id") or playlist_number
        name = data.get("name", "Unnamed")
        playlist_url = data.get("playlist_url", None)
        tracks = data.get("tracks", [])

        return cls(
            bot=bot,
            playlist_api=playlist_api,
            guild=guild,
            scope=scope,
            author=author,
            playlist_id=playlist_id,
            name=name,
            playlist_url=playlist_url,
            tracks=tracks,
        )

    async def save(self):
        """Saves a Playlist to SQL."""
        scope, scope_id = prepare_config_scope(self.bot, self.scope, self.author, self.guild)
        await self.playlist_api.upsert(
            scope,
            playlist_id=int(self.id),
            playlist_name=self.name,
            scope_id=scope_id,
            author_id=self.author,
            playlist_url=self.url,
            tracks=self.tracks,
        )


async def get_all_playlist_for_migration23(
    bot: Red,
    playlist_api: PlaylistWrapper,
    config: Config,
    scope: str,
    guild: Union[discord.Guild, int] = None,
) -> List[PlaylistCompat23]:
    """
    Gets all playlist for the specified scope.
    Parameters
    ----------
    bot: Red
        The Bot instance.
    playlist_api: PlaylistWrapper
        The Playlist API interface.
    config: Config
        The Audio cog Config instance.
    scope: str
        The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
    guild: discord.Guild
        The guild to get the playlist from if scope is GUILDPLAYLIST.
    Returns
    -------
    list
        A list of all playlists for the specified scope
    Raises
    ------
    `InvalidPlaylistScope`
        Passing a scope that is not supported.
    `MissingGuild`
        Trying to access the Guild scope without a guild.
    `MissingAuthor`
        Trying to access the User scope without an user id.
    """
    playlists = await config.custom(scope).all()
    if scope == PlaylistScope.GLOBAL.value:
        return [
            await PlaylistCompat23.from_json(
                bot,
                playlist_api,
                scope,
                playlist_number,
                playlist_data,
                guild=guild,
                author=int(playlist_data.get("author", 0)),
            )
            async for playlist_number, playlist_data in AsyncIter(playlists.items())
        ]
    elif scope == PlaylistScope.USER.value:
        return [
            await PlaylistCompat23.from_json(
                bot,
                playlist_api,
                scope,
                playlist_number,
                playlist_data,
                guild=guild,
                author=int(user_id),
            )
            async for user_id, scopedata in AsyncIter(playlists.items())
            async for playlist_number, playlist_data in AsyncIter(scopedata.items())
        ]
    else:
        return [
            await PlaylistCompat23.from_json(
                bot,
                playlist_api,
                scope,
                playlist_number,
                playlist_data,
                guild=int(guild_id),
                author=int(playlist_data.get("author", 0)),
            )
            async for guild_id, scopedata in AsyncIter(playlists.items())
            async for playlist_number, playlist_data in AsyncIter(scopedata.items())
        ]


async def get_playlist(
    playlist_number: int,
    scope: str,
    bot: Red,
    playlist_api: PlaylistWrapper,
    guild: Union[discord.Guild, int] = None,
    author: Union[discord.abc.User, int] = None,
) -> Playlist:
    """
    Gets the playlist with the associated playlist number.
    Parameters
    ----------
    playlist_number: int
        The playlist number for the playlist to get.
    playlist_api: PlaylistWrapper
        The Playlist API interface.
    scope: str
        The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
    guild: discord.Guild
        The guild to get the playlist from if scope is GUILDPLAYLIST.
    author: int
        The ID of the user to get the playlist from if scope is USERPLAYLIST.
    bot: Red
        The bot's instance.
    Returns
    -------
    Playlist
        The playlist associated with the playlist number.
    Raises
    ------
    `RuntimeError`
        If there is no playlist for the specified number.
    `InvalidPlaylistScope`
        Passing a scope that is not supported.
    `MissingGuild`
        Trying to access the Guild scope without a guild.
    `MissingAuthor`
        Trying to access the User scope without an user id.
    """
    scope_standard, scope_id = prepare_config_scope(bot, scope, author, guild)
    playlist_data = await playlist_api.fetch(scope_standard, playlist_number, scope_id)

    if not (playlist_data and playlist_data.playlist_id):
        raise RuntimeError(f"That playlist does not exist for the following scope: {scope}")
    return await Playlist.from_json(
        bot,
        playlist_api,
        scope_standard,
        playlist_number,
        playlist_data,
        guild=guild,
        author=author,
    )


async def get_all_playlist(
    scope: str,
    bot: Red,
    playlist_api: PlaylistWrapper,
    guild: Union[discord.Guild, int] = None,
    author: Union[discord.abc.User, int] = None,
    specified_user: bool = False,
) -> List[Playlist]:
    """
    Gets all playlist for the specified scope.
    Parameters
    ----------
    scope: str
        The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
    guild: discord.Guild
        The guild to get the playlist from if scope is GUILDPLAYLIST.
    author: int
        The ID of the user to get the playlist from if scope is USERPLAYLIST.
    bot: Red
        The bot's instance
    playlist_api: PlaylistWrapper
        The Playlist API interface.
    specified_user:bool
        Whether or not user ID was passed as an argparse.
    Returns
    -------
    list
        A list of all playlists for the specified scope
    Raises
    ------
    `InvalidPlaylistScope`
        Passing a scope that is not supported.
    `MissingGuild`
        Trying to access the Guild scope without a guild.
    `MissingAuthor`
        Trying to access the User scope without an user id.
    """
    scope_standard, scope_id = prepare_config_scope(bot, scope, author, guild)

    if specified_user:
        user_id = getattr(author, "id", author)
        playlists = await playlist_api.fetch_all(scope_standard, scope_id, author_id=user_id)
    else:
        playlists = await playlist_api.fetch_all(scope_standard, scope_id)

    playlist_list = []
    async for playlist in AsyncIter(playlists):
        playlist_list.append(
            await Playlist.from_json(
                bot,
                playlist_api,
                scope,
                playlist.playlist_id,
                playlist,
                guild=guild,
                author=author,
            )
        )
    return playlist_list


async def get_all_playlist_converter(
    scope: str,
    bot: Red,
    playlist_api: PlaylistWrapper,
    arg: str,
    guild: Union[discord.Guild, int] = None,
    author: Union[discord.abc.User, int] = None,
) -> List[Playlist]:
    """
    Gets all playlist for the specified scope.
    Parameters
    ----------
    scope: str
        The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
    guild: discord.Guild
        The guild to get the playlist from if scope is GUILDPLAYLIST.
    author: int
        The ID of the user to get the playlist from if scope is USERPLAYLIST.
    bot: Red
        The bot's instance
    arg:str
        The value to lookup.
    playlist_api: PlaylistWrapper
        The Playlist API interface.
    Returns
    -------
    list
        A list of all playlists for the specified scope
    Raises
    ------
    `InvalidPlaylistScope`
        Passing a scope that is not supported.
    `MissingGuild`
        Trying to access the Guild scope without a guild.
    `MissingAuthor`
        Trying to access the User scope without an user id.
    """
    scope_standard, scope_id = prepare_config_scope(bot, scope, author, guild)
    playlists = await playlist_api.fetch_all_converter(
        scope_standard, playlist_name=arg, playlist_id=arg
    )
    playlist_list = []
    async for playlist in AsyncIter(playlists):
        playlist_list.append(
            await Playlist.from_json(
                bot,
                playlist_api,
                scope,
                playlist.playlist_id,
                playlist,
                guild=guild,
                author=author,
            )
        )
    return playlist_list


async def create_playlist(
    ctx: commands.Context,
    playlist_api: PlaylistWrapper,
    scope: str,
    playlist_name: str,
    playlist_url: Optional[str] = None,
    tracks: Optional[List[MutableMapping]] = None,
    author: Optional[discord.User] = None,
    guild: Optional[discord.Guild] = None,
) -> Optional[Playlist]:
    """Creates a new Playlist.

    Parameters
    ----------
    ctx: commands.Context
        The context in which the play list is being created.
    scope: str
        The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
    playlist_name: str
        The name of the new playlist.
    playlist_url:str
        the url of the new playlist.
    tracks: List[MutableMapping]
        A list of tracks to add to the playlist.
    author: discord.User
        The Author of the playlist.
        If provided it will create a playlist under this user.
        This is only required when creating a playlist in User scope.
    guild: discord.Guild
        The guild to create this playlist under.
         This is only used when creating a playlist in the Guild scope
    playlist_api: PlaylistWrapper
        The Playlist API interface.

    Raises
    ------
    `InvalidPlaylistScope`
        Passing a scope that is not supported.
    `MissingGuild`
        Trying to access the Guild scope without a guild.
    `MissingAuthor`
        Trying to access the User scope without an user id.
    """

    playlist = Playlist(
        ctx.bot,
        playlist_api,
        scope,
        author.id if author else None,
        ctx.message.id,
        playlist_name,
        playlist_url,
        tracks,
        guild or ctx.guild,
    )
    await playlist.save()
    return playlist


async def reset_playlist(
    bot: Red,
    playlist_api: PlaylistWrapper,
    scope: str,
    guild: Union[discord.Guild, int] = None,
    author: Union[discord.abc.User, int] = None,
) -> None:
    """Wipes all playlists for the specified scope.

    Parameters
    ----------
    bot: Red
        The bot's instance
    scope: str
        The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
    guild: discord.Guild
        The guild to get the playlist from if scope is GUILDPLAYLIST.
    author: int
        The ID of the user to get the playlist from if scope is USERPLAYLIST.
    playlist_api: PlaylistWrapper
        The Playlist API interface.

    Raises
    ------
    `InvalidPlaylistScope`
        Passing a scope that is not supported.
    `MissingGuild`
        Trying to access the Guild scope without a guild.
    `MissingAuthor`
        Trying to access the User scope without an user id.
    """
    scope, scope_id = prepare_config_scope(bot, scope, author, guild)
    await playlist_api.drop(scope)
    await playlist_api.create_table()


async def delete_playlist(
    bot: Red,
    playlist_api: PlaylistWrapper,
    scope: str,
    playlist_id: Union[str, int],
    guild: discord.Guild,
    author: Union[discord.abc.User, int] = None,
) -> None:
    """Deletes the specified playlist.

    Parameters
    ----------
    bot: Red
        The bot's instance
    scope: str
        The custom config scope. One of 'GLOBALPLAYLIST', 'GUILDPLAYLIST' or 'USERPLAYLIST'.
    playlist_id: Union[str, int]
        The ID of the playlist.
    guild: discord.Guild
        The guild to get the playlist from if scope is GUILDPLAYLIST.
    author: int
        The ID of the user to get the playlist from if scope is USERPLAYLIST.
    playlist_api: PlaylistWrapper
        The Playlist API interface.

    Raises
    ------
    `InvalidPlaylistScope`
        Passing a scope that is not supported.
    `MissingGuild`
        Trying to access the Guild scope without a guild.
    `MissingAuthor`
        Trying to access the User scope without an user id.
    """
    scope, scope_id = prepare_config_scope(bot, scope, author, guild)
    await playlist_api.delete(scope, int(playlist_id), scope_id)



import weakref
from typing import Tuple, Type

from redbot.core.config import Config, _config_cache
from redbot.core._drivers import BaseDriver

__all__ = ("get_latest_confs", "migrate")

_retrieved = weakref.WeakSet()


def get_latest_confs() -> Tuple[Config, ...]:
    global _retrieved
    ret = set(_config_cache.values()) - set(_retrieved)
    _retrieved |= ret
    return tuple(ret)


async def migrate(cur_driver_cls: Type[BaseDriver], new_driver_cls: Type[BaseDriver]) -> None:
    """Migrate from one driver type to another."""
    # Get custom group data
    core_conf = Config.get_core_conf(allow_old=True)
    core_conf.init_custom("CUSTOM_GROUPS", 2)
    all_custom_group_data = await core_conf.custom("CUSTOM_GROUPS").all()

    await cur_driver_cls.migrate_to(new_driver_cls, all_custom_group_data)
